from __future__ import annotations

from math import pi, sqrt
from typing import TYPE_CHECKING

import numpy as np
from ase.units import Bohr, Ha

from gpaw.atom.shapefunc import shape_functions
from gpaw.core.arrays import DistributedArrays
from gpaw.core.atom_arrays import AtomArrays
from gpaw.core.uniform_grid import UGArray
from gpaw.setup import Setups
from gpaw.typing import Array1D, ArrayLike2D
from gpaw.utilities import pack

if TYPE_CHECKING:
    from gpaw.new.calculation import DFTCalculation


class ElectrostaticPotential:
    def __init__(self,
                 vHt_x: DistributedArrays,
                 W_aL: AtomArrays,
                 Q_aL: AtomArrays,
                 D_asii: AtomArrays,
                 fracpos_ac: ArrayLike2D,
                 setups: Setups):
        self.vHt_x = vHt_x
        self.W_aL = W_aL
        self.Q_aL = Q_aL
        self.D_asii = D_asii
        self.fracpos_ac = fracpos_ac
        self.setups = setups

        # Caching of interpolated pseudo-potential:
        self._grid_spacing = -1.0
        self._vHt_R: UGArray | None = None

    @classmethod
    def from_calculation(cls, calculation: DFTCalculation):
        density = calculation.state.density
        potential, W_aL = calculation.pot_calc.calculate(density)
        Q_aL = density.calculate_compensation_charge_coefficients()
        return cls(potential.vHt_x,
                   W_aL,
                   Q_aL,
                   density.D_asii,
                   calculation.fracpos_ac,
                   calculation.setups)

    def atomic_potentials(self) -> Array1D:
        W_aL = self.W_aL.gather(broadcast=True)
        return W_aL.data[::9] * (Ha / (4 * pi)**0.5)

    def atomic_corrections(self):
        """Calculate PAW correction to average electrostatic potential."""
        dEH_a = np.zeros(len(self.setups))
        for a, D_sii in self.D_asii.items():
            setup = self.setups[a]
            D_p = pack(D_sii.sum(0))
            dEH_a[a] = setup.dEH0 + setup.dEH_p @ D_p
        self.D_asii.comm.sum(dEH_a)
        return dEH_a * Ha * Bohr**3

    def pseudo_potential(self,
                         grid_spacing: float = 0.05,  # Ang
                         ) -> UGArray:
        return self._pseudo_potential(grid_spacing / Bohr).scaled(Bohr, Ha)

    def _pseudo_potential(self,
                          grid_spacing: float,  # Bohr
                          ) -> UGArray:
        if grid_spacing == self._grid_spacing:
            assert self._vHt_R is not None
            return self._vHt_R

        vHt_x = self.vHt_x
        if isinstance(self.vHt_x, UGArray):
            vHt_x = self.vHt_x.to_pbc_grid()
        grid = vHt_x.desc.uniform_grid_with_grid_spacing(grid_spacing / Bohr)
        self._vHt_R = vHt_x.interpolate(grid=grid)
        self._grid_spacing = grid_spacing
        return self._vHt_R

    def all_electron_potential(self,
                               grid_spacing: float = 0.05,  # Ang
                               rcgauss: float = 0.02,  # Ang
                               npoints: int = 200) -> UGArray:
        """Interpolate electrostatic potential.

        Return value in eV.

        ae: bool
            Add PAW correction to get the all-electron potential.
        rcgauss: float
            Width of gaussian (in Angstrom) used to represent the nuclear
            charge.
        """
        vHt_R = self._pseudo_potential(grid_spacing / Bohr)

        dv_a = []
        for a, D_sii in self.D_asii.items():
            setup = self.setups[a]
            c = setup.xc_correction
            rgd = c.rgd
            params = setup.data.shape_function.copy()
            params['lmax'] = 0
            ghat_g = shape_functions(rgd, **params)[0]
            Z_g = shape_functions(rgd, 'gauss', rcgauss, lmax=0)[0] * setup.Z
            D_p = pack(D_sii.sum(axis=0))
            D_q = D_p @ c.B_pqL[:, :, 0]
            dn_g = D_q @ (c.n_qg - c.nt_qg) * sqrt(4 * pi)
            dn_g += 4 * pi * (c.nc_g - c.nct_g)
            dn_g -= Z_g
            dn_g -= self.Q_aL[a][0] * ghat_g * sqrt(4 * pi)
            dv_g = rgd.poisson(dn_g) / sqrt(4 * pi)
            dv_g[1:] /= rgd.r_g[1:]
            dv_g[0] = dv_g[1]
            dv_g[-1] = 0.0
            dv_a.append([rgd.spline(dv_g, points=npoints)])

        dv_aR = vHt_R.desc.atom_centered_functions(dv_a, self.fracpos_ac)
        dv_aR.add_to(vHt_R)
        return vHt_R.scaled(Bohr, Ha)



import pytest
from ase.build import molecule
from gpaw import GPAW, PW
from gpaw.xc.rpa import RPACorrelation
from gpaw.hybrids.energy import non_self_consistent_energy as nsc_energy


@pytest.mark.rpa
@pytest.mark.response
def test_rpa_rpa_energy_N2(in_tmp_dir, scalapack):
    ecut = 25

    N2 = molecule('N2')
    N2.center(vacuum=2.0)

    calc = GPAW(mode=PW(force_complex_dtype=True),
                xc='PBE',
                parallel={'domain': 1},
                eigensolver='rmm-diis')
    N2.calc = calc
    E_n2_pbe = N2.get_potential_energy()

    E_n2_hf = nsc_energy(N2.calc, 'EXX')

    calc.diagonalize_full_hamiltonian(nbands=104, scalapack=True)
    calc.write('N2.gpw', mode='all')

    rpa = RPACorrelation('N2.gpw', nfrequencies=8, ecut=[ecut])
    E_n2_rpa = rpa.calculate()

    N = molecule('N')
    N.set_cell(N2.cell)
    N.center()

    calc = GPAW(mode=PW(force_complex_dtype=True),
                xc='PBE',
                parallel={'domain': 1},
                eigensolver='rmm-diis')
    N.calc = calc
    E_n_pbe = N.get_potential_energy()

    calc.diagonalize_full_hamiltonian(nbands=104, scalapack=True)
    calc.write('N.gpw', mode='all')

    E_n_hf = nsc_energy('N.gpw', 'EXX')

    rpa = RPACorrelation('N.gpw', nfrequencies=8, ecut=[ecut])
    E_n_rpa = rpa.calculate()

    print('Atomization energies:')
    print('PBE: ', E_n2_pbe - 2 * E_n_pbe)
    print('HF: ', E_n2_hf - 2 * E_n_hf)
    print('HF+RPA: ', E_n2_hf - 2 * E_n_hf + E_n2_rpa[0] - 2 * E_n_rpa[0])

    assert E_n2_rpa - 2 * E_n_rpa == pytest.approx(-1.68, abs=0.02)
    assert (E_n2_hf - 2 * E_n_hf) == pytest.approx(
        [-10.47, 3.03, 0, 0, -0.06, 2.79], abs=0.01)



r"""Module for calculating electron-phonon matrix.

Electron-phonon interaction::

                  __
                  \     l   +         +
        H      =   )   g   c   c   ( a   + a  ),
         el-ph    /_    ij  i   j     l     l
                 l,ij

where the electron phonon coupling is given by::

                      ______
             l       / hbar         ___
            g   =   /-------  < i | \ /  V   * e  | j > .
             ij   \/ 2 M w           'u   eff   l
                          l

Here, l denotes the vibrational mode, w_l and e_l is the frequency and
mass-scaled polarization vector, respectively, M is an effective mass, i, j are
electronic state indices and nabla_u denotes the gradient wrt atomic
displacements. The implementation supports calculations of the el-ph coupling
in both finite and periodic systems, i.e. expressed in a basis of molecular
orbitals or Bloch states.
"""
import numpy as np
from typing import Optional

from ase import Atoms
from ase.phonons import Phonons
import ase.units as units
from ase.utils.filecache import MultiFileJSONCache
from ase.utils.timing import timer, Timer

from gpaw.calculator import GPAW
from gpaw.mpi import world
from gpaw.typing import ArrayND

from .supercell import Supercell

OPTIMIZE = "optimal"


class ElectronPhononMatrix:
    """Class for containing the electron-phonon matrix"""

    def __init__(self, atoms: Atoms, supercell_cache: str, phonon,
                 load_sc_as_needed: bool = True) -> None:
        """Initialize with base class args and kwargs.

        Parameters
        ----------
        atoms: Atoms
            Primitive cell object
        supercell_cache: str
            Name of JSON cache containing supercell matrix
        phonon: str, dict, :class:`~ase.phonons.Phonons`
            Can be either name of phonon cache generated with
            electron-phonon DisplacementRunner or dictonary
            of arguments used in Phonons run or Phonons object.
        load_sc_as_needed: bool
            Load supercell matrix elements only as needed.
            Greatly reduces memory requirement for large systems,
            but introduces huge filesystem overhead
        """
        self.timer = Timer()

        self.atoms = atoms
        self.supercell_cache = MultiFileJSONCache(supercell_cache)
        self.R_cN = self._get_lattice_vectors()

        if load_sc_as_needed:
            self._yield_g_NNMM = self._yield_g_NNMM_as_needed
            self.g_xNNMM = None
        else:
            self.g_xsNNMM, _ = Supercell.load_supercell_matrix(supercell_cache)
            self._yield_g_NNMM = self._yield_g_NNMM_from_var

        self.timer.start("Read phonons")
        if isinstance(phonon, Phonons):
            self.phonon = phonon
        elif isinstance(phonon, str):
            info = MultiFileJSONCache(phonon)["info"]
            assert "dr_version" in info, "use valid cache created by elph"
            # our version of phonons
            self.phonon = Phonons(atoms, supercell=info["supercell"],
                                  name=phonon, delta=info["delta"],
                                  center_refcell=True)
        elif isinstance(phonon, dict):
            # this would need to be updated of Phonon defaults change
            supercell = phonon.get("supercell", (1, 1, 1))
            name = phonon.get("name", "phonon")
            delta = phonon.get("delta", 0.01)
            center_refcell = phonon.get("center_refcell", False)
            self.phonon = Phonons(atoms, None, supercell, name, delta,
                                  center_refcell)
        else:
            raise TypeError

        if self.phonon.D_N is None:
            self.phonon.read(symmetrize=10)
        self.timer.stop("Read phonons")

    def _yield_g_NNMM_as_needed(self, x, s):
        return self.supercell_cache[str(x)][s]

    def _yield_g_NNMM_from_var(self, x, s):
        return self.g_xsNNMM[x, s]

    def _get_lattice_vectors(self):
        """Recover lattice vectors of elph calculation"""
        supercell = self.supercell_cache["info"]["supercell"]
        ph = Phonons(self.atoms, supercell=supercell, center_refcell=True)
        return ph.compute_lattice_vectors()

    @classmethod
    def _gather_all_wfc(cls, wfs, s):
        """Return complete wave function on rank 0"""
        c_knM = np.zeros((wfs.kd.nbzkpts, wfs.bd.nbands, wfs.setups.nao),
                         dtype=complex)
        for k in range(wfs.kd.nbzkpts):
            for n in range(wfs.bd.nbands):
                c_knM[k, n] = wfs.get_wave_function_array(n, k, s, False)
        return c_knM

    @timer("Bloch matrix q k")
    def _bloch_matrix(self, var1: ArrayND, C2_nM: ArrayND,
                      k_c: ArrayND, q_c: ArrayND,
                      prefactor: bool, s: Optional[int] = None) -> ArrayND:
        """Calculates elph matrix entry for a given k and q.

        The first argument must either be
        C1_nM, the ket wavefunction at k_c
        OR
        or a preprocessed g_xNMn, where the ket side was taken care of.
        """
        if var1.ndim == 2:
            C1_nM = var1
            precalc = False
            assert s is not None
        elif var1.ndim == 4:
            g_xNMn = var1
            precalc = True
        else:
            raise ValueError("var1 must be C1_nM or g_xNMn")
        omega_ql, u_ql = self.phonon.band_structure([q_c], modes=True)
        u_l = u_ql[0]
        assert len(u_l.shape) == 3

        # Defining system sizes
        nmodes = u_l.shape[0]
        nbands = C2_nM.shape[0]
        nao = C2_nM.shape[1]
        ndisp = 3 * len(self.atoms)

        # Allocate array for couplings
        g_lnn = np.zeros((nmodes, nbands, nbands), dtype=complex)

        # Mass scaled polarization vectors
        u_lx = u_l.reshape(nmodes, ndisp)

        # Multiply phase factors
        phase_m = np.exp(2.0j * np.pi * np.einsum("i,im->m", k_c + q_c,
                                                  self.R_cN))
        if not precalc:
            phase_n = np.exp(-2.0j * np.pi * np.einsum("i,in->n", k_c,
                                                       self.R_cN))
        for x in range(ndisp):
            if not precalc:
                g_NNMM = self._yield_g_NNMM(x, s)
                assert nao == g_NNMM.shape[-1]
                # some of these things take a long time. make it fast
                with self.timer("g_MM"):
                    g_MM = np.einsum("mnop,m,n->op", g_NNMM, phase_m,
                                     phase_n, optimize=OPTIMIZE)
                    assert g_MM.shape[0] == g_MM.shape[1]
                with self.timer("g_nn"):
                    g_nn = np.dot(C2_nM.conj(), np.dot(g_MM, C1_nM.T))
                    # g_nn = np.einsum('no,op,mp->nm', C2_nM.conj(),
                    #                   g_MM, C1_nM)
            else:
                with self.timer("g_Mn"):
                    g_Mn = np.einsum("mon,m->on", g_xNMn[x], phase_m)
                with self.timer("g_nn"):
                    g_nn = np.dot(C2_nM.conj(), g_Mn)
            with self.timer("g_lnn"):
                # g_lnn += np.einsum('i,kl->ikl', u_lx[:, x], g_nn,
                #                   optimize=OPTIMIZE)
                g_lnn += np.multiply.outer(u_lx[:, x], g_nn)

        # Multiply prefactor sqrt(hbar / 2 * M * omega) in units of Bohr
        if prefactor:
            # potential BUG: M needs to be unit cell mass according to
            # some sources
            amu = units._amu  # atomic mass unit
            me = units._me  # electron mass
            g_lnn /= np.sqrt(2 * amu / me / units.Hartree *
                             omega_ql[0, :, np.newaxis, np.newaxis])
            # Convert to eV
            return g_lnn * units.Hartree  # eV
        else:
            return g_lnn * units.Hartree / units.Bohr  # eV / Ang

    @timer("g ket part")
    def _precalculate_ket(self, c_knM, kd, s: int):
        g_xNkMn = []
        phase_kn = np.exp(-2.0j * np.pi * np.einsum("ki,in->kn", kd.bzk_kc,
                                                    self.R_cN))
        for x in range(3 * len(self.atoms)):
            g_NNMM = self._yield_g_NNMM_as_needed(x, s)
            g_NkMM = np.einsum("mnop,kn->mkop", g_NNMM, phase_kn,
                               optimize=OPTIMIZE)
            g_NkMn = np.einsum("mkop,knp->mkon", g_NkMM, c_knM,
                               optimize=OPTIMIZE)

            g_xNkMn.append(g_NkMn)
        return np.array(g_xNkMn)

    def bloch_matrix(self, calc: GPAW, k_qc: ArrayND = None,
                     savetofile: bool = True, prefactor: bool = True,
                     accoustic: bool = True) -> ArrayND:
        r"""Calculate el-ph coupling in the Bloch basis for the electrons.

        This function calculates the electron-phonon coupling between the
        specified Bloch states, i.e.::

                      ______
            mnl      / hbar               ^
           g    =   /-------  < m k + q | e  . grad V  | n k >
            kq    \/ 2 M w                 ql        q
                          ql

        In case the ``prefactor=False`` is given, the bare matrix
        element (in units of eV / Ang) without the sqrt prefactor is returned.

        Parameters
        ----------
        calc: GPAW
            Converged calculator object containing the LCAO wavefuntions
            (don't use point group symmetry)
        k_qc: np.ndarray
            q-vectors of the phonons. Must only contain values comenserate
            with k-point sampling of calculator. Default: all kpoints used.
        savetofile: bool
            If true (default), saves matrix to gsqklnn.npy
        prefactor: bool
            if false, don't multiply with sqrt prefactor (Default: True)
        accoustic: bool
            if True, for 3 accoustic modes set g=0 for q=0 (Default: True)
        """
        kd = calc.wfs.kd
        assert kd.nbzkpts == kd.nibzkpts, "Elph matrix requires FULL BZ"

        wfs = calc.wfs
        if k_qc is None:
            k_qc = kd.get_bz_q_points(first=True)
        elif not isinstance(k_qc, np.ndarray):
            k_qc = np.array(k_qc)
        assert k_qc.ndim == 2

        g_sqklnn = np.zeros([wfs.nspins, k_qc.shape[0], kd.nbzkpts,
                             3 * len(self.atoms), wfs.bd.nbands,
                             wfs.bd.nbands], dtype=complex)

        for s in range(wfs.nspins):
            # Collect all wfcs on rank 0
            with self.timer("Gather wavefunctions to root"):
                c_knM = self._gather_all_wfc(wfs, s)

            # precalculate k (ket) of g
            g_xNkMn = self._precalculate_ket(c_knM, kd, s)

            for q, q_c in enumerate(k_qc):
                # Find indices of k+q for the k-points
                kplusq_k = kd.find_k_plus_q(q_c)  # works on FBZ
                # Note: calculations require use of FULL BZ,
                # so NO symmetry
                print("Spin {}/{}; q-point {}/{}".format(
                    s + 1, wfs.nspins, q + 1, len(k_qc)))

                for k in range(kd.nbzkpts):
                    k_c = kd.bzk_kc[k]
                    kplusq_c = k_c + q_c
                    kplusq_c -= kplusq_c.round()
                    # print(kplusq_c, kd.bzk_kc[kplusq_k[k]])
                    assert np.allclose(kplusq_c, kd.bzk_kc[kplusq_k[k]])
                    ckplusq_nM = c_knM[kplusq_k[k]]
                    g_lnn = self._bloch_matrix(g_xNkMn[:, :, k], ckplusq_nM,
                                               k_c, q_c, prefactor)
                    if np.allclose(q_c, [0.0, 0.0, 0.0]) and accoustic:
                        g_lnn[0:3] = 0.0
                    g_sqklnn[s, q, k] += g_lnn

        if world.rank == 0 and savetofile:
            np.save("gsqklnn.npy", g_sqklnn)

        return g_sqklnn

    def __del__(self):
        if world.rank == 0:
            self.timer.write()

#   def lcao_matrix(self, u_l, omega_l):
#         """Calculate the el-ph coupling in the electronic LCAO basis.

#         For now, only works for Gamma-point phonons.

#         This method is not tested.

#         Parameters
#         ----------
#         u_l: ndarray
#             Mass-scaled polarization vectors (in units of 1 / sqrt(amu)) of
#             the phonons.
#         omega_l: ndarray
#             Vibrational frequencies in eV.
#         """

#         # Supercell matrix (Hartree / Bohr)
#         assert self.g_xsNNMM is not None, "Load supercell matrix."
#         assert self.g_xsNNMM.shape[2:4] == (1, 1)
#         g_xsMM = self.g_xsNNMM[:, :, 0, 0, :, :]
#         # Number of atomic orbitals
#         # nao = g_xMM.shape[-1]
#         # Number of phonon modes
#         nmodes = u_l.shape[0]

#         #
#         u_lx = u_l.reshape(nmodes, 3 * len(self.atoms))
#         # np.dot uses second to last index of second array
#         g_lsMM = np.dot(u_lx, g_xsMM.transpose(2, 0, 1, 3))

#         # Multiply prefactor sqrt(hbar / 2 * M * omega) in units of Bohr
#         amu = units._amu  # atomic mass unit
#         me = units._me   # electron mass
#         g_lsMM /= np.sqrt(2 * amu / me / units.Hartree *
#                           omega_l[:, :, np.newaxis, np.newaxis])
#         # Convert to eV
#         g_lsMM *= units.Hartree

#         return g_lsMM



"""BSSE: Basis Set Superposition Error module.

Defines a Setup-like class which has no properties that change anything,
except for an atomic basis set."""

import numpy as np
from ase.data import atomic_numbers

from gpaw.setup import BaseSetup, LocalCorrectionVar
from gpaw.spline import Spline
from gpaw.utilities import min_locfun_radius


# Some splines are mandatory,
# but should then be zero to avoid affecting things
zero_function = Spline(0, min_locfun_radius, [0.0, 0.0, 0.0])

# Some operations fail horribly if the splines are zero, due to weird
# divisions and assumptions that various quantities are nonzero
#
# We'll use a function which is almost zero for these things
nonzero_function = Spline(0, min_locfun_radius, [0.0, 1.0e-12, 0.0])  # XXX


class GhostSetup(BaseSetup):
    def __init__(self, basis, data):
        self.symbol = data.symbol
        self.data = data
        self.basis_functions_J = basis.tosplines()
        self.pseudo_partial_waves_j = None  # XXX ?

        self.basis = basis
        self.nao = sum([2 * phit.get_angular_momentum_number() + 1
                        for phit in self.basis_functions_J])
        self.filename = None
        self.fingerprint = None
        self.type = 'ghost'

        self.Z = 0
        self.Nv = 0
        self.Nc = 0

        self.ni = 1
        self.pt_j = [zero_function]
        self.wg_lg = None
        self.g_lg = None

        self.Nct = 1e-12  # XXX XXX XXX XXX
        self.nct = nonzero_function  # XXXXXX
        self.lmax = 0
        self.xc_correction = None
        self.ghat_l = [nonzero_function] * (self.lmax + 1)  # XXXXXX
        self.rcgauss = 1e12  # XXX XXX XXX XXX
        self.vbar = zero_function

        self.Delta0 = 0.0
        self.Delta_pL = np.zeros((1, self.lmax + 1))

        self.E = 0.0
        self.Kc = 0.0
        self.M = 0.0
        self.M_p = np.zeros(1)
        self.M_pp = np.zeros((1, 1))
        self.M_wpp = {}
        self.K_p = np.zeros(1)
        self.MB = 0.0
        self.MB_p = np.zeros(1)
        self.dO_ii = np.zeros((1, 1))
        self.f_j = [0.0]
        self.n_j = [0]
        self.l_j = [0]
        self.l_orb_J = [0]
        self.nj = 1
        self.lq = None  # XXXX

        self.rcutfilter = None
        self.rcore = None
        self.N0_p = np.zeros(1)
        self.nabla_iiv = np.zeros((self.ni, self.ni, 3))
        self.rxnabla_iiv = np.zeros((self.ni, self.ni, 3))
        self.phicorehole_g = None
        self.rgd = None
        self.rcut_j = [0.5]
        self.tauct = None
        self.Delta_iiL = None
        self.B_ii = None
        self.dC_ii = None
        self.X_p = None
        self.X_wp = {}
        self.X_pg = None
        self.ExxC = None
        self.ExxC_w = {}
        self.X_gamma = None
        self.dEH0 = 0.0
        self.dEH_p = np.zeros(1)
        self.extra_xc_data = {}
        self.local_corr = LocalCorrectionVar(None)
        self._Mg_pp = None
        self._gamma = None


class GhostSetupData:
    def __init__(self, symbol):
        self.chemsymbol = symbol
        self.symbol = symbol + '.ghost'
        self.Z = atomic_numbers[symbol]

    def build(self, xcfunc, lmax, basis, filter=None):
        if basis is None:
            raise ValueError('Loading partial waves not supported right now')
        setup = GhostSetup(basis, self)
        return setup

    def print_info(self, text, _setup):
        text('Ghost setup for %s' % self.chemsymbol)



import numpy as np
from time import ctime

from ase.units import Hartree

from gpaw.utilities.blas import mmmx

from gpaw.response import ResponseGroundStateAdapter, ResponseContext, timer
from gpaw.response.frequencies import ComplexFrequencyDescriptor
from gpaw.response.pw_parallelization import PlaneWaveBlockDistributor
from gpaw.response.matrix_elements import NewPairDensityCalculator
from gpaw.response.pair_integrator import PairFunctionIntegrator
from gpaw.response.pair_transitions import PairTransitions
from gpaw.response.pair_functions import SingleQPWDescriptor, Chi


class ChiKSCalculator(PairFunctionIntegrator):
    r"""Calculator class for the four-component Kohn-Sham susceptibility tensor
    of collinear systems in absence of spin-orbit coupling,
    see [PRB 103, 245110 (2021)]:
                              __  __   __
                           1  \   \    \
    χ_KS,GG'^μν(q,ω+iη) =  ‾  /   /    /   σ^μ_ss' σ^ν_s's (f_nks - f_n'k+qs')
                           V  ‾‾  ‾‾   ‾‾
                              k   n,n' s,s'
                                        n_nks,n'k+qs'(G+q) n_n'k+qs',nks(-G'-q)
                                      x ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
                                            ħω - (ε_n'k+qs' - ε_nks) + iħη

    where the matrix elements

    n_nks,n'k+qs'(G+q) = <nks| e^-i(G+q)r |n'k+qs'>_V0

    are the unit cell normalized plane-wave pair densities of each transition.
    """

    def __init__(self, gs: ResponseGroundStateAdapter, context=None,
                 nblocks=1,
                 ecut=50, gammacentered=False,
                 nbands=None,
                 bandsummation='pairwise',
                 **kwargs):
        """Contruct the ChiKSCalculator

        Parameters
        ----------
        gs : ResponseGroundStateAdapter
        context : ResponseContext
        nblocks : int
            Distribute the chiks_zGG array into nblocks (where nblocks is a
            divisor of context.comm.size)
        ecut : float (or None)
            Plane-wave cutoff in eV
        gammacentered : bool
            Center the grid of plane waves around the Γ-point (or the q-vector)
        nbands : int
            Number of bands to include in the sum over states
        bandsummation : str
            Band summation strategy (does not change the result, but can affect
            the run-time).
            'pairwise': sum over pairs of bands
            'double': double sum over band indices.
        kwargs : see gpaw.response.pair_integrator.PairFunctionIntegrator
        """
        if context is None:
            context = ResponseContext()
        assert isinstance(context, ResponseContext)

        super().__init__(gs, context, nblocks=nblocks, **kwargs)

        self.ecut = None if ecut is None else ecut / Hartree  # eV to Hartree
        self.gammacentered = gammacentered
        self.nbands = nbands
        self.bandsummation = bandsummation

        self.pair_density_calc = NewPairDensityCalculator(gs, context)

    def calculate(self, spincomponent, q_c, zd) -> Chi:
        r"""Calculate χ_KS,GG'^μν(q,z), where z = ω + iη

        Parameters
        ----------
        spincomponent : str
            Spin component (μν) of the Kohn-Sham susceptibility.
            Currently, '00', 'uu', 'dd', '+-' and '-+' are implemented.
        q_c : list or np.array
            Wave vector in relative coordinates
        zd : ComplexFrequencyDescriptor
            Complex frequencies z to evaluate χ_KS,GG'^μν(q,z) at.
        """
        return self._calculate(*self._set_up_internals(spincomponent, q_c, zd))

    def _set_up_internals(self, spincomponent, q_c, zd,
                          distribution='GZg'):
        r"""Set up internal data objects to calculate χ_KS."""
        assert isinstance(zd, ComplexFrequencyDescriptor)

        # Set up the internal plane-wave descriptor
        qpdi = self.get_pw_descriptor(q_c, internal=True)

        # Prepare to sum over bands and spins
        transitions = self.get_band_and_spin_transitions(
            spincomponent, nbands=self.nbands,
            bandsummation=self.bandsummation)

        self.context.print(self.get_info_string(
            qpdi, len(zd), spincomponent, len(transitions)))

        # Create data structure
        chiks = self.create_chiks(spincomponent, qpdi, zd, distribution)

        return chiks, transitions

    def _calculate(self, chiks: Chi, transitions: PairTransitions):
        r"""Integrate χ_KS according to the specified transitions."""
        self.context.print('Initializing pair density PAW corrections')
        self.pair_density_calc.initialize_paw_corrections(chiks.qpd)

        # Perform the actual integration
        analyzer = self._integrate(chiks, transitions)

        # Symmetrize chiks according to the symmetries of the ground state
        self.symmetrize(chiks, analyzer)

        # Map to standard output format
        chiks = self.post_process(chiks)

        return chiks

    def get_pw_descriptor(self, q_c, internal=False):
        """Get plane-wave descriptor for the wave vector q_c.

        Parameters
        ----------
        q_c : list or ndarray
            Wave vector in relative coordinates
        internal : bool
            When using symmetries, the actual calculation of chiks must happen
            using a q-centered plane wave basis. If internal==True, as it is by
            default, the internal plane wave basis (used in the integration of
            chiks.array) is returned, otherwise the external descriptor is
            returned, corresponding to the requested chiks.
        """
        q_c = np.asarray(q_c, dtype=float)
        gd = self.gs.gd

        # Update to internal basis, if needed
        if internal and self.gammacentered and not self.disable_symmetries:
            # In order to make use of the symmetries of the system to reduce
            # the k-point integration, the internal code assumes a plane wave
            # basis which is centered at q in reciprocal space.
            gammacentered = False
            # If we want to compute the pair function on a plane wave grid
            # which is effectively centered in the gamma point instead of q, we
            # need to extend the internal ecut such that the q-centered grid
            # encompasses all reciprocal lattice points inside the gamma-
            # centered sphere.
            # The reduction to the global gamma-centered basis will then be
            # carried out as a post processing step.

            # Compute the extended internal ecut
            B_cv = 2.0 * np.pi * gd.icell_cv  # Reciprocal lattice vectors
            q_v = q_c @ B_cv
            ecut = get_ecut_to_encompass_centered_sphere(q_v, self.ecut)
        else:
            gammacentered = self.gammacentered
            ecut = self.ecut

        qpd = SingleQPWDescriptor.from_q(q_c, ecut, gd,
                                         gammacentered=gammacentered)

        return qpd

    def create_chiks(self, spincomponent, qpd, zd, distribution):
        """Create a new Chi object to be integrated."""
        assert distribution in ['GZg', 'ZgG']
        blockdist = PlaneWaveBlockDistributor(self.context.comm,
                                              self.blockcomm,
                                              self.intrablockcomm)
        return Chi(spincomponent, qpd, zd,
                   blockdist, distribution=distribution)

    @timer('Add integrand to chiks')
    def add_integrand(self, kptpair, weight, chiks):
        r"""Use the NewPairDensityCalculator object to calculate the integrand
        for all relevant transitions of the given k-point pair, k -> k + q.

        Depending on the bandsummation parameter, the integrand of the
        collinear four-component Kohn-Sham susceptibility tensor (in the
        absence of spin-orbit coupling) is calculated as:

        bandsummation: double

                   __
                   \  σ^μ_ss' σ^ν_s's (f_nks - f_n'k's')
        (...)_k =  /  ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾ n_kt(G+q) n_kt^*(G'+q)
                   ‾‾      ħz - (ε_n'k's' - ε_nks)
                   t

        where n_kt(G+q) = n_nks,n'k+qs'(G+q) and

        bandsummation: pairwise

                    __ /
                    \  | σ^μ_ss' σ^ν_s's (f_nks - f_n'k's')
        (...)_k =   /  | ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
                    ‾‾ |      ħz - (ε_n'k's' - ε_nks)
                    t  \
                                                       \
                    σ^μ_s's σ^ν_ss' (f_nks - f_n'k's') |
           - δ_n'>n ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾ | n_kt(G+q) n_kt^*(G'+q)
                         ħz + (ε_n'k's' - ε_nks)       |
                                                       /

        The integrand is added to the output array chiks_x multiplied with the
        supplied kptpair integral weight.
        """
        # Calculate the pair densities n_kt(G+q)
        pair_density = self.pair_density_calc(kptpair, chiks.qpd)

        # Extract the temporal ingredients from the KohnShamKPointPair
        transitions = kptpair.transitions  # transition indices (n,s)->(n',s')
        df_t = kptpair.df_t  # (f_n'k's' - f_nks)
        deps_t = kptpair.deps_t  # (ε_n'k's' - ε_nks)

        # Calculate the temporal part of the integrand
        if chiks.spincomponent == '00' and self.gs.nspins == 1:
            weight = 2 * weight
        x_Zt = get_temporal_part(chiks.spincomponent, chiks.zd.hz_z,
                                 transitions, df_t, deps_t,
                                 self.bandsummation)

        self._add_integrand(pair_density, x_Zt, weight, chiks)

    def _add_integrand(self, pair_density, x_Zt, weight, chiks):
        r"""Add the integrand to chiks.

        This entail performing a sum of transition t and an outer product
        in the pair density plane wave components G and G',
                    __
                    \
        (...)_k =   /  x_t^μν(ħz) n_kt(G+q) n_kt^*(G'+q)
                    ‾‾
                    t

        where x_t^μν(ħz) is the temporal part of χ_KS,GG'^μν(q,ω+iη).
        """
        if chiks.distribution == 'ZgG':
            self._add_integrand_ZgG(pair_density, x_Zt, weight, chiks)
        elif chiks.distribution == 'GZg':
            self._add_integrand_GZg(pair_density, x_Zt, weight, chiks)
        else:
            raise ValueError(f'Invalid distribution {chiks.distribution}')

    def _add_integrand_ZgG(self, pair_density, x_Zt, weight, chiks):
        """Add integrand in ZgG distribution.

        Z = global complex frequency index
        g = distributed G plane wave index
        G = global G' plane wave index
        """
        chiks_ZgG = chiks.array
        myslice = chiks.blocks1d.myslice

        with self.context.timer('Set up ncc and xn'):
            # Multiply the temporal part with the k-point integration weight
            x_Zt *= weight

            # Set up n_kt^*(G'+q)
            n_tG = pair_density.get_global_array()
            ncc_tG = n_tG.conj()

            # Set up x_t^μν(ħz) n_kt(G+q)
            n_gt = np.ascontiguousarray(n_tG[:, myslice].T)
            xn_Zgt = x_Zt[:, np.newaxis, :] * n_gt[np.newaxis, :, :]

        with self.context.timer('Perform sum over t-transitions of xn * ncc'):
            for xn_gt, chiks_gG in zip(xn_Zgt, chiks_ZgG):
                mmmx(1.0, xn_gt, 'N', ncc_tG, 'N', 1.0, chiks_gG)  # slow step

    def _add_integrand_GZg(self, pair_density, x_Zt, weight, chiks):
        """Add integrand in GZg distribution.

        G = global G' plane wave index
        Z = global complex frequency index
        g = distributed G plane wave index
        """
        chiks_GZg = chiks.array
        myslice = chiks.blocks1d.myslice

        with self.context.timer('Set up ncc and xn'):
            # Multiply the temporal part with the k-point integration weight
            x_tZ = np.ascontiguousarray(weight * x_Zt.T)

            # Set up n_kt^*(G'+q)
            n_tG = pair_density.get_global_array()
            n_Gt = np.ascontiguousarray(n_tG.T)
            ncc_Gt = n_Gt.conj()

            # Set up x_t^μν(ħz) n_kt(G+q)
            n_tg = n_tG[:, myslice]
            xn_tZg = x_tZ[:, :, np.newaxis] * n_tg[:, np.newaxis, :]

        with self.context.timer('Perform sum over t-transitions of ncc * xn'):
            mmmx(1.0, ncc_Gt, 'N', xn_tZg, 'N', 1.0, chiks_GZg)  # slow step

    @timer('Symmetrizing chiks')
    def symmetrize(self, chiks, analyzer):
        """Symmetrize chiks_zGG."""
        chiks_ZgG = chiks.array_with_view('ZgG')

        # Distribute over frequencies
        nz = len(chiks.zd)
        tmp_zGG = chiks.blockdist.distribute_as(chiks_ZgG, nz, 'zGG')
        analyzer.symmetrize_zGG(tmp_zGG)
        # Distribute over plane waves
        chiks_ZgG[:] = chiks.blockdist.distribute_as(tmp_zGG, nz, 'ZgG')

    @timer('Post processing')
    def post_process(self, chiks):
        """Cast a calculated chiks into a fixed output format."""
        if chiks.distribution != 'ZgG':
            # Always output chiks with distribution 'ZgG'
            chiks = chiks.copy_with_distribution('ZgG')

        if self.gammacentered and not self.disable_symmetries:
            # Reduce the q-centered plane-wave basis used internally to the
            # gammacentered basis
            assert not chiks.qpd.gammacentered  # Internal qpd
            qpd = self.get_pw_descriptor(chiks.q_c)  # External qpd
            chiks = chiks.copy_with_reduced_pd(qpd)

        return chiks

    def get_info_string(self, qpd, nz, spincomponent, nt):
        r"""Get information about the χ_KS,GG'^μν(q,z) calculation"""
        from gpaw.utilities.memory import maxrss

        q_c = qpd.q_c
        ecut = qpd.ecut * Hartree
        Asize = nz * qpd.ngmax**2 * 16. / 1024**2 / self.blockcomm.size
        cmem = maxrss() / 1024**2

        isl = ['',
               'Setting up a Kohn-Sham susceptibility calculation with:',
               f'    Spin component: {spincomponent}',
               f'    q_c: [{q_c[0]}, {q_c[1]}, {q_c[2]}]',
               f'    Number of frequency points: {nz}',
               self.get_band_and_transitions_info_string(self.nbands, nt),
               '',
               self.get_basic_info_string(),
               '',
               'Plane-wave basis of the Kohn-Sham susceptibility:',
               f'    Planewave cutoff: {ecut}',
               f'    Number of planewaves: {qpd.ngmax}',
               '    Memory estimates:',
               f'        A_zGG: {Asize} M / cpu',
               f'        Memory usage before allocation: {cmem} M / cpu',
               '',
               f'{ctime()}']

        return '\n'.join(isl)


def get_ecut_to_encompass_centered_sphere(q_v, ecut):
    """Calculate the minimal ecut which results in a q-centered plane wave
    basis containing all the reciprocal lattice vectors G, which lie inside a
    specific gamma-centered sphere:

    |G|^2 < 2 * ecut
    """
    q = np.linalg.norm(q_v)
    ecut += q * (np.sqrt(2 * ecut) + q / 2)

    return ecut


def get_temporal_part(spincomponent, hz_z,
                      transitions, df_t, deps_t,
                      bandsummation):
    """Get the temporal part of a (causal linear) susceptibility, x_t^μν(ħz).
    """
    _get_temporal_part = create_get_temporal_part(bandsummation)
    return _get_temporal_part(spincomponent, hz_z,
                              transitions, df_t, deps_t)


def create_get_temporal_part(bandsummation):
    """Creator component, deciding how to calculate the temporal part"""
    if bandsummation == 'double':
        return get_double_temporal_part
    elif bandsummation == 'pairwise':
        return get_pairwise_temporal_part
    raise ValueError(bandsummation)


def get_double_temporal_part(spincomponent, hz_z,
                             transitions, df_t, deps_t):
    r"""Get:

                 σ^μ_ss' σ^ν_s's (f_nks - f_n'k's')
    x_t^μν(ħz) = ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
                      ħz - (ε_n'k's' - ε_nks)
    """
    # Get the right spin components
    s1_t, s2_t = transitions.get_spin_indices()
    scomps_t = get_smat_components(spincomponent, s1_t, s2_t)
    # Calculate nominator
    nom_t = - scomps_t * df_t  # df = f2 - f1
    # Calculate denominator
    denom_wt = hz_z[:, np.newaxis] - deps_t[np.newaxis, :]  # de = e2 - e1

    regularize_intraband_transitions(denom_wt, transitions, deps_t)

    return nom_t[np.newaxis, :] / denom_wt


def get_pairwise_temporal_part(spincomponent, hz_z,
                               transitions, df_t, deps_t):
    r"""Get:

                 /
                 | σ^μ_ss' σ^ν_s's (f_nks - f_n'k's')
    x_t^μν(ħz) = | ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
                 |      ħz - (ε_n'k's' - ε_nks)
                 \
                                                               \
                            σ^μ_s's σ^ν_ss' (f_nks - f_n'k's') |
                   - δ_n'>n ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾ |
                                 ħz + (ε_n'k's' - ε_nks)       |
                                                               /
    """
    n1_t, n2_t, s1_t, s2_t = transitions.get_band_and_spin_indices()
    # Kroenecker delta
    delta_t = np.ones(len(n1_t))
    delta_t[n2_t <= n1_t] = 0
    # Get the right spin components
    scomps1_t = get_smat_components(spincomponent, s1_t, s2_t)
    scomps2_t = get_smat_components(spincomponent, s2_t, s1_t)
    # Calculate nominators
    nom1_t = - scomps1_t * df_t  # df = f2 - f1
    nom2_t = - delta_t * scomps2_t * df_t
    # Calculate denominators
    denom1_wt = hz_z[:, np.newaxis] - deps_t[np.newaxis, :]  # de = e2 - e1
    denom2_wt = hz_z[:, np.newaxis] + deps_t[np.newaxis, :]

    regularize_intraband_transitions(denom1_wt, transitions, deps_t)
    regularize_intraband_transitions(denom2_wt, transitions, deps_t)

    return nom1_t[np.newaxis, :] / denom1_wt\
        - nom2_t[np.newaxis, :] / denom2_wt


def regularize_intraband_transitions(denom_wt, transitions, deps_t):
    """Regularize the denominator of the temporal part in case of degeneracy.

    If the q-vector connects two symmetrically equivalent k-points inside a
    band, the occupation differences vanish and we regularize the denominator.

    NB: In principle there *should* be a contribution from the intraband
    transitions, but this is left for future work for now."""
    intraband_t = transitions.get_intraband_mask()
    degenerate_t = np.abs(deps_t) < 1e-8

    denom_wt[:, intraband_t & degenerate_t] = 1.


def get_smat_components(spincomponent, s1_t, s2_t):
    """For s1=s and s2=s', get:
    smu_ss' snu_s's
    """
    smatmu = smat(spincomponent[0])
    smatnu = smat(spincomponent[1])

    return smatmu[s1_t, s2_t] * smatnu[s2_t, s1_t]


def smat(spinrot):
    if spinrot == '0':
        return np.array([[1, 0], [0, 1]])
    elif spinrot == 'u':
        return np.array([[1, 0], [0, 0]])
    elif spinrot == 'd':
        return np.array([[0, 0], [0, 1]])
    elif spinrot == '-':
        return np.array([[0, 0], [1, 0]])
    elif spinrot == '+':
        return np.array([[0, 1], [0, 0]])
    elif spinrot == 'z':
        return np.array([[1, 0], [0, -1]])
    else:
        raise ValueError(spinrot)



import numpy as np
from gpaw.atom.radialgd import EquidistantRadialGridDescriptor


def test_smoothness():
    """Make sure pseudize_smooth() produces the most smooth function."""
    rgd = EquidistantRadialGridDescriptor(0.05, 200)
    r = rgd.r_g

    a = 3.0
    l = 0
    ecut = 10
    Gcut = (2 * ecut)**0.5
    gc = 20

    f0 = np.exp(-a * r**2)
    f1, _ = rgd.pseudize_normalized(f0, gc, l, 4)
    f2, _ = rgd.pseudize(f0, gc, l, 4)
    f3, _ = rgd.pseudize_smooth(f0, gc, l, 4, Gcut)

    # Calculate weight of large wave vectors (G > Gcut):
    weights = []
    for f in [f0, f1, f2, f3]:
        x, ft = rgd.fft(f * r, l=l)
        weights.append(((x * ft)[x > Gcut]**2).sum())

    print(weights)
    assert (np.diff(weights) < 0).all()



"""Simple tight-binding add-on for GPAWs LCAO module."""

from math import pi

import numpy as np
import scipy.linalg as sla

import ase.units as units

from gpaw.utilities.tools import tri2full


class TightBinding:
    """Simple class for tight-binding calculations."""

    def __init__(self, atoms, calc):
        """Init with ``Atoms`` and a converged LCAO calculation."""

        # Store and extract useful attributes from the calculator
        self.atoms = atoms
        self.calc = calc
        self.kd = calc.wfs.kd
        wfs = calc.wfs
        kd = wfs.kd

        # Matrix size
        self.nao = wfs.setups.nao
        # K-point info
        self.gamma = kd.gamma
        if self.gamma:
            self.Nk_c = (1, 1, 1)
        else:
            self.Nk_c = tuple(kd.N_c)

        self.ibzk_kc = kd.ibzk_kc
        self.ibzk_qc = kd.ibzk_qc
        self.bzk_kc = kd.bzk_kc

        # Symmetry
        self.symmetry = kd.symmetry
        if self.symmetry.point_group:
            raise NotImplementedError("Only time-reversal symmetry supported.")

        # Lattice vectors and number of repetitions
        self.R_cN = None
        self.N_c = None

        # Init with default number of real-space cells
        self.set_num_cells()

    def set_num_cells(self, N_c=None):
        """Set number of real-space cells to use.

        Parameters
        ----------
        N_c: tuple or ndarray
            Number of unit cells in each direction of the basis vectors.

        """

        if N_c is None:
            self.N_c = tuple(self.Nk_c)
        else:
            self.N_c = tuple(N_c)

        if np.any(np.asarray(self.Nk_c) < np.asarray(self.N_c)):
            print("WARNING: insufficient k-point sampling.")

        # Lattice vectors
        R_cN = np.indices(self.N_c).reshape(3, -1)
        N_c = np.array(self.N_c)[:, np.newaxis]
        R_cN += N_c // 2
        R_cN %= N_c
        R_cN -= N_c // 2
        self.R_cN = R_cN

    def lattice_vectors(self):
        """Return real-space lattice vectors."""

        return self.R_cN

    def bloch_to_real_space(self, A_qxMM, R_c=None):
        """Transform quantity from Bloch to real-space representation.

        Parameters
        ----------
        A_qxMM: ndarray
            Bloch representation of matrix. May be parallelized over k-points.
        R_cN: ndarray
            Cell vectors for which the real-space matrices will be calculated
            and returned.

        """

        # Include all cells per default
        if R_c is None:
            R_Nc = self.R_cN.transpose()
        else:
            R_Nc = [R_c]

        # Real-space quantities
        A_NxMM = []
        # Reshape input array
        shape = A_qxMM.shape
        A_qx = A_qxMM.reshape(shape[0], -1)

        # Fourier transform to real-space
        for R_c in R_Nc:
            # Evaluate fourier sum
            phase_q = np.exp(2.j * pi * np.dot(self.ibzk_qc, R_c))
            A_x = np.sum(phase_q[:, np.newaxis] * A_qx, axis=0)
            self.kd.comm.sum(A_x)
            A_xMM = A_x.reshape(shape[1:])

            # Time-reversal symmetry
            if not len(self.ibzk_kc) == len(self.bzk_kc):
                # Broadcast Gamma component
                gamma = np.where(abs(self.ibzk_kc).sum(axis=1) < 1e-13)[0]
                if gamma.size > 0:
                    rank, myu = self.kd.get_rank_and_index(gamma[0])

                    if self.kd.comm.rank == rank:
                        A0_xMM = A_qxMM[myu]
                    else:
                        A0_xMM = np.zeros_like(A_xMM)
                    self.kd.comm.broadcast(A0_xMM, rank)
                else:
                    A0_xMM = 0.
                # Add conjugate and subtract double counted Gamma component
                A_xMM += A_xMM.conj() - A0_xMM

            A_xMM /= np.prod(self.Nk_c)

            try:
                assert np.all(np.abs(A_xMM.imag) < 1e-10)
            except AssertionError:
                raise ValueError("MAX Im(A_MM): % .2e" %
                                 np.amax(np.abs(A_xMM.imag)))

            A_NxMM.append(A_xMM.real)

        return np.array(A_NxMM)

    def h_and_s(self):
        """Return LCAO Hamiltonian and overlap matrix in real-space."""

        # Extract Bloch Hamiltonian and overlap matrix
        H_kMM = []
        S_kMM = []

        h = self.calc.hamiltonian
        wfs = self.calc.wfs
        kpt_u = wfs.kpt_u

        for kpt in kpt_u:
            H_MM = wfs.eigensolver.calculate_hamiltonian_matrix(h, wfs, kpt)
            S_MM = wfs.S_qMM[kpt.q]
            # XXX Converting to full matrices here
            tri2full(H_MM)
            tri2full(S_MM)
            H_kMM.append(H_MM)
            S_kMM.append(S_MM)

        # Convert to arrays
        H_kMM = np.array(H_kMM)
        S_kMM = np.array(S_kMM)

        H_NMM = self.bloch_to_real_space(H_kMM)
        S_NMM = self.bloch_to_real_space(S_kMM)

        return H_NMM, S_NMM

    def band_structure(self, path_kc, blochstates=False):
        """Calculate dispersion along a path in the Brillouin zone.

        Parameters
        ----------
        path_kc: ndarray
            List of k-point coordinates (in units of the reciprocal lattice
            vectors) specifying the path in the Brillouin zone for which the
            dynamical matrix will be calculated.
        blochstates: bool
            Return LCAO expansion coefficients when True (default: False).

        """

        # Real-space matrices
        self.H_NMM, self.S_NMM = self.h_and_s()

        assert self.H_NMM is not None
        assert self.S_NMM is not None

        # Lattice vectors
        R_cN = self.R_cN

        # Lists for eigenvalues and eigenvectors along path
        eps_kn = []
        psi_kn = []

        for k_c in path_kc:
            # Evaluate fourier sum
            phase_N = np.exp(-2.j * pi * np.dot(k_c, R_cN))
            H_MM = np.sum(phase_N[:, np.newaxis, np.newaxis] * self.H_NMM,
                          axis=0)
            S_MM = np.sum(phase_N[:, np.newaxis, np.newaxis] * self.S_NMM,
                          axis=0)

            if blochstates:
                eps_n, c_Mn = sla.eigh(H_MM, S_MM)
                # Sort eigenmodes according to increasing eigenvalues
                c_nM = c_Mn[:, eps_n.argsort()].transpose()
                psi_kn.append(c_nM)
            else:
                eps_n = sla.eigvalsh(H_MM, S_MM)

            # Sort eigenvalues in increasing order
            eps_n.sort()
            eps_kn.append(eps_n)

        # Convert to eV
        eps_kn = np.array(eps_kn) * units.Hartree

        if blochstates:
            return eps_kn, np.array(psi_kn)

        return eps_kn



import argparse
from typing import List, Union

import numpy as np
from ase import Atoms
from gpaw.new.ase_interface import ASECalculator, GPAW
from gpaw.point_groups import SymmetryChecker, point_group_names
from gpaw.typing import Array1D, Array3D


class CubeCalc:
    """Wrap cube-file in a calculator."""
    def __init__(self, function: Array3D, atoms: Atoms):
        self.function = function
        self.atoms = atoms

    def get_pseudo_wave_function(self,
                                 band: int,
                                 spin: int) -> Array3D:
        return self.function

    def get_eigenvalues(self, spin: int) -> Array1D:
        return np.zeros(1)

    def get_number_of_spins(self):
        return 1


def main(argv: List[str] = None) -> None:
    parser = argparse.ArgumentParser(
        prog='python3 -m gpaw.point_groups',
        description='Analyse point-group of atoms and wave-functions.')
    add = parser.add_argument
    add('pg', metavar='point-group', choices=point_group_names,
        help='Name of point-group: C2, C2v, C3v, D2d, D3h, D5, D5h, '
        'Ico, Ih, Oh, Td or Th.')
    add('file', metavar='input-file',
        help='Cube-file, gpw-file or something else with atoms in it.')
    add('-c', '--center', help='Center specified as one or more atoms.  '
        'Use chemical symbols or sequence numbers.')
    add('-r', '--radius', default=2.5,
        help='Cutoff radius (in Å) used for wave function overlaps.')
    add('-b', '--bands', default=':', metavar='N1:N2',
        help='Band range.')
    add('-a', '--axes', default='',
        help='Example: "-a z=x,x=-y".')
    args = parser.parse_intermixed_args(argv)

    calc: Union[None, ASECalculator, CubeCalc]

    if args.file.endswith('.gpw'):
        calc = GPAW(args.file)
        atoms = calc.atoms
        n1, n2 = (int(x) if x else 0 for x in args.bands.split(':'))
    elif args.file.endswith('.cube'):
        from ase.io.cube import read_cube
        with open(args.file) as fd:
            dct = read_cube(fd)
        calc = CubeCalc(dct['data'], dct['atoms'])
        atoms = dct['atoms']
        n1 = 0
        n2 = 1
    else:
        from ase.io import read
        atoms = read(args.file)
        calc = None

    if args.center:
        symbols = set(args.center.split(','))
        center = np.zeros(3)
        n = 0
        for a, (symbol, position) in enumerate(zip(atoms.symbols,
                                                   atoms.positions)):
            if symbol in symbols or str(a) in symbols:
                center += position
                n += 1
        center /= n
    else:
        center = atoms.cell.sum(0) / 2
    print('Center:', center, f'(atoms: {n})')

    radius = float(args.radius)

    kwargs = {}
    for axis in args.axes.split(',') if args.axes else []:
        axis1, axis2 = axis.split('=')
        kwargs[axis1] = axis2

    checker = SymmetryChecker(args.pg, center, radius, **kwargs)

    ok = checker.check_atoms(atoms)
    print(f'{args.pg}-symmetry:', 'Yes' if ok else 'No')

    if calc:
        nspins = calc.get_number_of_spins()
        for spin in range(nspins):
            if nspins == 2:
                print('Spin', ['up', 'down'][spin])
            checker.check_calculation(calc, n1, n2, spin=spin)



import numpy as np
import pytest
from ase import Atoms
from ase.phonons import Phonons

from gpaw import GPAW
from gpaw.elph.electronphonon import ElectronPhononCoupling
from gpaw.mpi import world

pytestmark = pytest.mark.skipif(world.size > 2,
                                reason='world.size > 2')


@pytest.mark.elph
def test_electronphonon(in_tmp_dir):
    a = 0.90
    atoms = Atoms('H',
                  cell=np.diag([a, 2.1, 2.1]),
                  positions=[[0, 0, 0]],
                  pbc=(1, 0, 0))

    atoms.center()
    supercell = (2, 1, 1)
    parameters = {'mode': 'lcao',
                  'kpts': {'size': (2, 1, 1), 'gamma': True},
                  'txt': None,
                  'basis': 'dzp',
                  'symmetry': {'point_group': False},
                  'parallel': {'domain': 1},
                  'xc': 'PBE'}
    elph_calc = GPAW(**parameters)
    atoms.calc = elph_calc
    atoms.get_potential_energy()
    gamma_bands = elph_calc.wfs.kpt_u[0].C_nM

    elph = ElectronPhononCoupling(atoms, elph_calc, supercell=supercell,
                                  name='elph+ph', calculate_forces=True)
    elph.run()

    elph.set_lcao_calculator(elph_calc)
    elph.calculate_supercell_matrix()

    ph = Phonons(atoms=atoms, name='elph+ph', supercell=supercell, calc=None)
    ph.read()
    kpts = [[0, 0, 0]]
    frequencies, modes = ph.band_structure(kpts, modes=True)

    c_kn = np.array([[gamma_bands[0]]])
    _ = elph.bloch_matrix(c_kn=c_kn, kpts=kpts, qpts=kpts, u_ql=modes)



import numpy as np

from ase.units import Hartree, Bohr

from ase.io.ulm import Reader
from gpaw.io import Writer
from gpaw.external import ConstantElectricField
from gpaw.lcaotddft.hamiltonian import KickHamiltonian
from gpaw.lcaotddft.utilities import collect_MM
from gpaw.lcaotddft.utilities import distribute_nM
from gpaw.lcaotddft.utilities import read_uMM
from gpaw.lcaotddft.utilities import write_uMM
from gpaw.lcaotddft.utilities import read_uX, write_uX
from gpaw.utilities.scalapack import \
    pblas_simple_gemm, pblas_simple_hemm, scalapack_tri2full
from gpaw.utilities.tools import tri2full


def gauss_ij(energy_i, energy_j, sigma):
    denergy_ij = energy_i[:, np.newaxis] - energy_j[np.newaxis, :]
    norm = 1.0 / (sigma * np.sqrt(2 * np.pi))
    return norm * np.exp(-0.5 * denergy_ij**2 / sigma**2)


def get_bfs_maps(calc):
    # Construct maps
    # a_M: M -> atom index a
    # l_M: M -> angular momentum l
    a_M = []
    l_M = []
    M = 0
    for a, sphere in enumerate(calc.wfs.basis_functions.sphere_a):
        for j, spline in enumerate(sphere.spline_j):
            l = spline.get_angular_momentum_number()
            for _ in range(2 * l + 1):
                a_M.append(a)
                l_M.append(l)
                M += 1
    a_M = np.array(a_M)
    l_M = np.array(l_M)
    return a_M, l_M


class KohnShamDecomposition(object):
    version = 1
    ulmtag = 'KSD'
    readwrite_attrs = ['fermilevel', 'only_ia', 'w_p', 'f_p', 'ia_p',
                       'P_p', 'dm_vp', 'a_M', 'l_M']

    def __init__(self, paw=None, filename=None):
        self.filename = filename
        self.has_initialized = False
        self.reader = None
        if paw is not None:
            self.world = paw.world
            self.log = paw.log
            self.ksl = paw.wfs.ksl
            self.kd = paw.wfs.kd
            self.bd = paw.wfs.bd
            self.kpt_u = paw.wfs.kpt_u
            self.density = paw.density
            self.comm = paw.comms['K']

            if len(paw.wfs.kpt_u) > 1:
                raise RuntimeError('K-points are not fully supported')

        if filename is not None:
            self.read(filename)
            return

    def initialize(self, paw, min_occdiff=1e-3, only_ia=True):
        if self.has_initialized:
            return
        paw.initialize_positions()
        # paw.set_positions()

        assert self.bd.nbands == self.ksl.nao
        self.only_ia = only_ia

        if not self.ksl.using_blacs and self.bd.comm.size > 1:
            raise RuntimeError('Band parallelization without scalapack '
                               'is not supported')

        if self.kd.gamma:
            self.C0_dtype = float
        else:
            self.C0_dtype = complex

        # Take quantities
        self.fermilevel = paw.wfs.fermi_level
        self.S_uMM = []
        self.C0_unM = []
        self.eig_un = []
        self.occ_un = []
        for kpt in paw.wfs.kpt_u:
            S_MM = kpt.S_MM
            assert np.max(np.absolute(S_MM.imag)) == 0.0
            S_MM = np.ascontiguousarray(S_MM.real)
            if self.ksl.using_blacs:
                scalapack_tri2full(self.ksl.mmdescriptor, S_MM)
            self.S_uMM.append(S_MM)

            C_nM = kpt.C_nM
            if self.C0_dtype == float:
                assert np.max(np.absolute(C_nM.imag)) == 0.0
                C_nM = np.ascontiguousarray(C_nM.real)
            C_nM = distribute_nM(self.ksl, C_nM)
            self.C0_unM.append(C_nM)

            eig_n = paw.wfs.collect_eigenvalues(kpt.k, kpt.s)
            occ_n = paw.wfs.collect_occupations(kpt.k, kpt.s)
            self.eig_un.append(eig_n)
            self.occ_un.append(occ_n)

        self.a_M, self.l_M = get_bfs_maps(paw)
        self.atoms = paw.atoms

        # TODO: do the rest of the function with K-points

        # Construct p = (i, a) pairs
        u = 0
        eig_n = self.eig_un[u]
        occ_n = self.occ_un[u]
        C0_nM = self.C0_unM[u]

        if self.comm.rank == 0:
            Nn = self.bd.nbands

            f_p = []
            w_p = []
            i_p = []
            a_p = []
            ia_p = []
            i0 = 0
            for i in range(i0, Nn):
                if only_ia:
                    a0 = i + 1
                else:
                    a0 = 0
                for a in range(a0, Nn):
                    f = occ_n[i] - occ_n[a]
                    if only_ia and f < min_occdiff:
                        continue
                    w = eig_n[a] - eig_n[i]
                    f_p.append(f)
                    w_p.append(w)
                    i_p.append(i)
                    a_p.append(a)
                    ia_p.append((i, a))
            f_p = np.array(f_p)
            w_p = np.array(w_p)
            i_p = np.array(i_p, dtype=int)
            a_p = np.array(a_p, dtype=int)
            ia_p = np.array(ia_p, dtype=int)

            # Sort according to energy difference
            p_s = np.argsort(w_p)
            f_p = f_p[p_s]
            w_p = w_p[p_s]
            i_p = i_p[p_s]
            a_p = a_p[p_s]
            ia_p = ia_p[p_s]

            Np = len(f_p)
            P_p = []
            for p in range(Np):
                P = np.ravel_multi_index(ia_p[p], (Nn, Nn))
                P_p.append(P)
            P_p = np.array(P_p)

            dm_vp = np.empty((3, Np), dtype=float)

        for v in range(3):
            direction = np.zeros(3, dtype=float)
            direction[v] = 1.0
            cef = ConstantElectricField(Hartree / Bohr, direction)
            kick_hamiltonian = KickHamiltonian(paw.hamiltonian, paw.density,
                                               cef)
            dm_MM = paw.wfs.eigensolver.calculate_hamiltonian_matrix(
                kick_hamiltonian, paw.wfs, paw.wfs.kpt_u[u],
                add_kinetic=False, root=-1)

            if self.ksl.using_blacs:
                tmp_nM = self.ksl.mmdescriptor.zeros(dtype=C0_nM.dtype)
                pblas_simple_hemm(self.ksl.mmdescriptor,
                                  self.ksl.mmdescriptor,
                                  self.ksl.mmdescriptor,
                                  dm_MM, C0_nM.conj(), tmp_nM,
                                  side='R', uplo='L')
                dm_nn = self.ksl.mmdescriptor.zeros(dtype=C0_nM.dtype)
                pblas_simple_gemm(self.ksl.mmdescriptor,
                                  self.ksl.mmdescriptor,
                                  self.ksl.mmdescriptor,
                                  tmp_nM, C0_nM, dm_nn, transb='T')
            else:
                tri2full(dm_MM)
                dm_nn = np.dot(C0_nM.conj(), np.dot(dm_MM, C0_nM.T))

            dm_nn = collect_MM(self.ksl, dm_nn)
            if self.comm.rank == 0:
                dm_P = dm_nn.ravel()
                dm_p = dm_P[P_p]
                dm_vp[v] = dm_p

        if self.comm.rank == 0:
            self.w_p = w_p
            self.f_p = f_p
            self.ia_p = ia_p
            self.P_p = P_p
            self.dm_vp = dm_vp

        self.has_initialized = True

    def write(self, filename):
        from ase.io.trajectory import write_atoms

        self.log('%s: Writing to %s' % (self.__class__.__name__, filename))
        writer = Writer(filename, self.world, mode='w',
                        tag=self.__class__.ulmtag)
        writer.write(version=self.__class__.version)

        write_atoms(writer.child('atoms'), self.atoms)

        writer.write(ha=Hartree)
        write_uMM(self.kd, self.ksl, writer, 'S_uMM', self.S_uMM)
        write_uMM(self.kd, self.ksl, writer, 'C0_unM', self.C0_unM)
        write_uX(self.kd, self.ksl.block_comm, writer, 'eig_un', self.eig_un)
        write_uX(self.kd, self.ksl.block_comm, writer, 'occ_un', self.occ_un)

        if self.comm.rank == 0:
            for arg in self.readwrite_attrs:
                writer.write(arg, getattr(self, arg))

        writer.close()

    def read(self, filename):
        self.reader = Reader(filename)
        tag = self.reader.get_tag()
        if tag != self.__class__.ulmtag:
            raise RuntimeError('Unknown tag %s' % tag)
        self.version = self.reader.version

        # Do lazy reading in __getattr__ only if/when
        # the variables are required
        self.has_initialized = True

    def __getattr__(self, attr):
        if attr in ['S_uMM', 'C0_unM']:
            val = read_uMM(self.kpt_u, self.ksl, self.reader, attr)
            setattr(self, attr, val)
            return val
        if attr in ['eig_un', 'occ_un']:
            val = read_uX(self.kpt_u, self.reader, attr)
            setattr(self, attr, val)
            return val
        if attr in ['C0S_unM']:
            C0S_unM = []
            for u, kpt in enumerate(self.kpt_u):
                C0_nM = self.C0_unM[u]
                S_MM = self.S_uMM[u]
                if self.ksl.using_blacs:
                    C0S_nM = self.ksl.mmdescriptor.zeros(dtype=C0_nM.dtype)
                    pblas_simple_hemm(self.ksl.mmdescriptor,
                                      self.ksl.mmdescriptor,
                                      self.ksl.mmdescriptor,
                                      S_MM, C0_nM, C0S_nM,
                                      side='R', uplo='L')
                else:
                    C0S_nM = np.dot(C0_nM, S_MM)
                C0S_unM.append(C0S_nM)
            setattr(self, attr, C0S_unM)
            return C0S_unM
        if attr in ['weight_Mn']:
            assert self.world.size == 1
            C2_nM = np.absolute(self.C0_unM[0])**2
            val = C2_nM.T / np.sum(C2_nM, axis=1)
            setattr(self, attr, val)
            return val

        try:
            val = getattr(self.reader, attr)
            if attr == 'atoms':
                from ase.io.trajectory import read_atoms
                val = read_atoms(val)
            setattr(self, attr, val)
            return val
        except (KeyError, AttributeError):
            pass

        raise AttributeError('Attribute %s not defined in version %s' %
                             (repr(attr), repr(self.version)))

    def distribute(self, comm):
        self.comm = comm
        N = comm.size
        self.Np = len(self.P_p)
        self.Nq = int(np.ceil(self.Np / float(N)))
        self.NQ = self.Nq * N
        self.w_q = self.distribute_p(self.w_p)
        self.f_q = self.distribute_p(self.f_p)
        self.dm_vq = self.distribute_xp(self.dm_vp)

    def distribute_p(self, a_p, a_q=None, root=0):
        if a_q is None:
            a_q = np.zeros(self.Nq, dtype=a_p.dtype)
        if self.comm.rank == root:
            a_Q = np.append(a_p, np.zeros(self.NQ - self.Np, dtype=a_p.dtype))
        else:
            a_Q = None
        self.comm.scatter(a_Q, a_q, root)
        return a_q

    def collect_q(self, a_q, root=0):
        if self.comm.rank == root:
            a_Q = np.zeros(self.NQ, dtype=a_q.dtype)
        else:
            a_Q = None
        self.comm.gather(a_q, root, a_Q)
        if self.comm.rank == root:
            a_p = a_Q[:self.Np]
        else:
            a_p = None
        return a_p

    def distribute_xp(self, a_xp):
        Nx = a_xp.shape[0]
        a_xq = np.zeros((Nx, self.Nq), dtype=a_xp.dtype)
        for x in range(Nx):
            self.distribute_p(a_xp[x], a_xq[x])
        return a_xq

    def transform(self, rho_uMM, broadcast=False):
        assert len(rho_uMM) == 1, 'K-points not implemented'
        u = 0
        rho_MM = np.ascontiguousarray(rho_uMM[u])
        C0S_nM = self.C0S_unM[u].astype(rho_MM.dtype, copy=True)
        # KS decomposition
        if self.ksl.using_blacs:
            tmp_nM = self.ksl.mmdescriptor.zeros(dtype=rho_MM.dtype)
            pblas_simple_gemm(self.ksl.mmdescriptor,
                              self.ksl.mmdescriptor,
                              self.ksl.mmdescriptor,
                              C0S_nM, rho_MM, tmp_nM)
            rho_nn = self.ksl.mmdescriptor.zeros(dtype=rho_MM.dtype)
            pblas_simple_gemm(self.ksl.mmdescriptor,
                              self.ksl.mmdescriptor,
                              self.ksl.mmdescriptor,
                              tmp_nM, C0S_nM, rho_nn, transb='C')
        else:
            rho_nn = np.dot(np.dot(C0S_nM, rho_MM), C0S_nM.T.conj())

        rho_nn = collect_MM(self.ksl, rho_nn)
        if self.comm.rank == 0:
            rho_P = rho_nn.ravel()
            # Remove de-excitation terms
            rho_p = rho_P[self.P_p]
            if self.only_ia:
                rho_p *= 2
        else:
            rho_p = None

        if broadcast:
            if self.comm.rank != 0:
                rho_p = np.zeros_like(self.P_p, dtype=rho_MM.dtype)
            self.comm.broadcast(rho_p, 0)
        rho_up = [rho_p]
        return rho_up

    def ialims(self):
        i_p = self.ia_p[:, 0]
        a_p = self.ia_p[:, 1]
        imin = np.min(i_p)
        imax = np.max(i_p)
        amin = np.min(a_p)
        amax = np.max(a_p)
        return imin, imax, amin, amax

    def M_p_to_M_ia(self, M_p):
        return self.M_ia_from_M_p(M_p)

    def M_ia_from_M_p(self, M_p):
        imin, imax, amin, amax = self.ialims()
        M_ia = np.zeros((imax - imin + 1, amax - amin + 1), dtype=M_p.dtype)
        for M, (i, a) in zip(M_p, self.ia_p):
            M_ia[i - imin, a - amin] = M
        return M_ia

    def plot_matrix(self, M_p):
        import matplotlib.pyplot as plt
        M_ia = self.M_ia_from_M_p(M_p)
        plt.imshow(M_ia, interpolation='none')
        plt.xlabel('a')
        plt.ylabel('i')

    def get_dipole_moment_contributions(self, rho_up):
        assert len(rho_up) == 1, 'K-points not implemented'
        u = 0
        rho_p = rho_up[u]
        dmrho_vp = - self.dm_vp * rho_p
        return dmrho_vp

    def get_dipole_moment(self, rho_up):
        assert len(rho_up) == 1, 'K-points not implemented'
        u = 0
        rho_p = rho_up[u]
        dm_v = - np.dot(self.dm_vp, rho_p)
        return dm_v

    def get_density(self, wfs, rho_up, density='comp'):
        from gpaw.lcaotddft.densitymatrix import get_density

        if self.ksl.using_blacs:
            raise NotImplementedError('Scalapack is not supported')

        density_type = density
        assert len(rho_up) == 1, 'K-points not implemented'
        u = 0
        rho_p = rho_up[u]
        C0_nM = self.C0_unM[u]

        rho_ia = self.M_ia_from_M_p(rho_p)
        imin, imax, amin, amax = self.ialims()
        C0_iM = C0_nM[imin:(imax + 1)]
        C0_aM = C0_nM[amin:(amax + 1)]

        rho_MM = np.dot(C0_iM.T, np.dot(rho_ia, C0_aM.conj()))
        rho_MM = 0.5 * (rho_MM + rho_MM.T)

        return get_density(rho_MM, wfs, self.density, density_type, u)

    def get_contributions_table(self, weight_p, minweight=0.01,
                                zero_fermilevel=True):
        assert weight_p.dtype == float
        u = 0  # TODO

        absweight_p = np.absolute(weight_p)
        tot_weight = weight_p.sum()
        propweight_p = weight_p / tot_weight * 100
        tot_propweight = propweight_p.sum()
        rest_weight = tot_weight
        rest_propweight = tot_propweight
        eig_n = self.eig_un[u].copy()
        if zero_fermilevel:
            eig_n -= self.fermilevel

        txt = ''
        txt += ('# %6s %4s(%8s)    %4s(%8s)  %12s %14s %8s\n' %
                ('p', 'i', 'eV', 'a', 'eV', 'Ediff (eV)', 'weight', '%'))
        p_s = np.argsort(absweight_p)[::-1]
        for s, p in enumerate(p_s):
            i, a = self.ia_p[p]
            if absweight_p[p] < minweight:
                break
            txt += ('  %6s %4d(%8.3f) -> %4d(%8.3f): %12.4f %14.4f %8.1f\n' %
                    (p, i, eig_n[i] * Hartree, a, eig_n[a] * Hartree,
                     self.w_p[p] * Hartree, weight_p[p], propweight_p[p]))
            rest_weight -= weight_p[p]
            rest_propweight -= propweight_p[p]
        txt += ('  %39s: %12s %+14.4f %8.1f\n' %
                ('rest', '', rest_weight, rest_propweight))
        txt += ('  %39s: %12s %+14.4f %8.1f\n' %
                ('total', '', tot_weight, tot_propweight))
        return txt

    def plot_TCM(self, weight_p, energy_o, energy_u, sigma,
                 zero_fermilevel=True, vmax='80%'):
        from gpaw.lcaotddft.tcm import TCMPlotter
        plotter = TCMPlotter(self, energy_o, energy_u, sigma, zero_fermilevel)
        ax_tcm = plotter.plot_TCM(weight_p, vmax)
        ax_occ_dos, ax_unocc_dos = plotter.plot_DOS()
        return ax_tcm, ax_occ_dos, ax_unocc_dos

    def get_TCM(self, weight_p, eig_n, energy_o, energy_u, sigma):
        flt_p = self.filter_by_x_ia(eig_n, energy_o, energy_u, 8 * sigma)
        weight_f = weight_p[flt_p]
        G_fo = gauss_ij(eig_n[self.ia_p[flt_p, 0]], energy_o, sigma)
        G_fu = gauss_ij(eig_n[self.ia_p[flt_p, 1]], energy_u, sigma)
        tcm_ou = np.dot(G_fo.T * weight_f, G_fu)
        return tcm_ou

    def get_DOS(self, eig_n, energy_o, energy_u, sigma):
        return self.get_weighted_DOS(1, eig_n, energy_o, energy_u, sigma)

    def get_weighted_DOS(self, weight_n, eig_n, energy_o, energy_u, sigma):
        if not isinstance(weight_n, np.ndarray):
            # Assume float
            weight_n = weight_n * np.ones_like(eig_n)
        G_on = gauss_ij(energy_o, eig_n, sigma)
        G_un = gauss_ij(energy_u, eig_n, sigma)
        dos_o = np.dot(G_on, weight_n)
        dos_u = np.dot(G_un, weight_n)
        return dos_o, dos_u

    def get_weight_n_by_l(self, l):
        if isinstance(l, int):
            weight_n = np.sum(self.weight_Mn[self.l_M == l], axis=0)
        else:
            weight_n = np.sum([self.get_weight_n_by_l(l_) for l_ in l],
                              axis=0)
        return weight_n

    def get_weight_n_by_a(self, a):
        if isinstance(a, int):
            weight_n = np.sum(self.weight_Mn[self.a_M == a], axis=0)
        else:
            weight_n = np.sum([self.get_weight_n_by_a(a_) for a_ in a],
                              axis=0)
        return weight_n

    def get_distribution_i(self, weight_p, energy_e, sigma,
                           zero_fermilevel=True):
        eig_n, fermilevel = self.get_eig_n(zero_fermilevel)
        flt_p = self.filter_by_x_i(eig_n, energy_e, 8 * sigma)
        weight_f = weight_p[flt_p]
        G_fe = gauss_ij(eig_n[self.ia_p[flt_p, 0]], energy_e, sigma)
        dist_e = np.dot(G_fe.T, weight_f)
        return dist_e

    def get_distribution_a(self, weight_p, energy_e, sigma,
                           zero_fermilevel=True):
        eig_n, fermilevel = self.get_eig_n(zero_fermilevel)
        flt_p = self.filter_by_x_a(eig_n, energy_e, 8 * sigma)
        weight_f = weight_p[flt_p]
        G_fe = gauss_ij(eig_n[self.ia_p[flt_p, 1]], energy_e, sigma)
        dist_e = np.dot(G_fe.T, weight_f)
        return dist_e

    def get_distribution_ia(self, weight_p, energy_o, energy_u, sigma,
                            zero_fermilevel=True):
        """
        Filter both i and a spaces as in TCM.

        """
        eig_n, fermilevel = self.get_eig_n(zero_fermilevel)
        flt_p = self.filter_by_x_ia(eig_n, energy_o, energy_u, 8 * sigma)
        weight_f = weight_p[flt_p]
        G_fo = gauss_ij(eig_n[self.ia_p[flt_p, 0]], energy_o, sigma)
        dist_o = np.dot(G_fo.T, weight_f)
        G_fu = gauss_ij(eig_n[self.ia_p[flt_p, 1]], energy_u, sigma)
        dist_u = np.dot(G_fu.T, weight_f)
        return dist_o, dist_u

    def get_distribution(self, weight_p, energy_e, sigma):
        w_p = self.w_p * Hartree
        flt_p = self.filter_by_x_p(w_p, energy_e, 8 * sigma)
        weight_f = weight_p[flt_p]
        G_fe = gauss_ij(w_p[flt_p], energy_e, sigma)
        dist_e = np.dot(G_fe.T, weight_f)
        return dist_e

    def get_eig_n(self, zero_fermilevel=True):
        u = 0  # TODO
        eig_n = self.eig_un[u].copy()
        if zero_fermilevel:
            eig_n -= self.fermilevel
            fermilevel = 0.0
        else:
            fermilevel = self.fermilevel
        eig_n *= Hartree
        fermilevel *= Hartree
        return eig_n, fermilevel

    def filter_by_x_p(self, x_p, energy_e, buf):
        flt_p = np.logical_and((energy_e[0] - buf) <= x_p,
                               x_p <= (energy_e[-1] + buf))
        return flt_p

    def filter_by_x_i(self, x_n, energy_e, buf):
        return self.filter_by_x_p(x_n[self.ia_p[:, 0]], energy_e, buf)

    def filter_by_x_a(self, x_n, energy_e, buf):
        return self.filter_by_x_p(x_n[self.ia_p[:, 1]], energy_e, buf)

    def filter_by_x_ia(self, x_n, energy_o, energy_u, buf):
        flti_p = self.filter_by_x_i(x_n, energy_o, buf)
        flta_p = self.filter_by_x_a(x_n, energy_u, buf)
        flt_p = np.logical_and(flti_p, flta_p)
        return flt_p

    def __del__(self):
        if self.reader is not None:
            self.reader.close()



import pytest
from ase import Atom, Atoms
from gpaw import GPAW, Davidson, Mixer
from gpaw.test import equal
from gpaw.xc.hybrid import HybridXC


@pytest.mark.mgga
@pytest.mark.libxc
def test_xc_atomize(in_tmp_dir, gpaw_new):
    def xc(name):
        return {'name': name, 'stencil': 1}

    a = 6.  # Size of unit cell (Angstrom)
    c = a / 2
    # Hydrogen atom:
    atom = Atoms([Atom('H', (c, c, c), magmom=1)],
                 cell=(a, a, a), pbc=False)

    # gpaw calculator:
    calc = GPAW(mode='fd',
                gpts=(32, 32, 32),
                nbands=1,
                xc=xc('PBE'),
                txt='H.txt',
                eigensolver=Davidson(12),
                mixer=Mixer(0.5, 5),
                parallel=dict(kpt=1),
                convergence=dict(eigenstates=3.3e-8))
    atom.calc = calc

    e1 = atom.get_potential_energy()
    de1t = calc.get_xc_difference(xc('TPSS'))
    de1m = calc.get_xc_difference(xc('M06-L'))
    if not gpaw_new:
        de1x = calc.get_xc_difference(
            HybridXC('EXX', stencil=1, finegrid=True))
        de1xb = calc.get_xc_difference(
            HybridXC('EXX', stencil=1, finegrid=False))

    # Hydrogen molecule:
    d = 0.74  # Experimental bond length
    molecule = Atoms([Atom('H', (c - d / 2, c, c)),
                      Atom('H', (c + d / 2, c, c))],
                     cell=(a, a, a), pbc=False)

    molecule.calc = calc.new(txt='H2.txt')
    e2 = molecule.get_potential_energy()
    de2t = molecule.calc.get_xc_difference(xc('TPSS'))
    de2m = molecule.calc.get_xc_difference(xc('M06-L'))

    print('hydrogen atom energy:     %5.2f eV' % e1)
    print('hydrogen molecule energy: %5.2f eV' % e2)
    print('atomization energy:       %5.2f eV' % (2 * e1 - e2))
    print('atomization energy  TPSS: %5.2f eV' %
          (2 * (e1 + de1t) - (e2 + de2t)))
    print('atomization energy  M06-L: %5.2f eV' %
          (2 * (e1 + de1m) - (e2 + de2m)))
    PBETPSSdifference = (2 * e1 - e2) - (2 * (e1 + de1t) - (e2 + de2t))
    PBEM06Ldifference = (2 * e1 - e2) - (2 * (e1 + de1m) - (e2 + de2m))
    print(PBETPSSdifference)
    print(PBEM06Ldifference)
    # TPSS value is from JCP 120 (15) 6898, 2004
    # e.g. Table VII: DE(PBE - TPSS) = (104.6-112.9)*kcal/mol
    # EXX value is from PRL 77, 3865 (1996)
    equal(PBETPSSdifference, -0.3599, 0.04)
    equal(PBEM06Ldifference, -0.169, 0.01)

    energy_tolerance = 0.002
    equal(e1, -1.081638, energy_tolerance)
    equal(e2, -6.726356, energy_tolerance)

    if not gpaw_new:
        de2x = molecule.calc.get_xc_difference(
            HybridXC('EXX', stencil=1, finegrid=True))
        de2xb = molecule.calc.get_xc_difference(
            HybridXC('EXX', stencil=1, finegrid=False))

        print('atomization energy   EXX: %5.2f eV' %
              (2 * (e1 + de1x) - (e2 + de2x)))
        print('atomization energy   EXX: %5.2f eV' %
              (2 * (e1 + de1xb) - (e2 + de2xb)))
        PBEEXXdifference = (2 * e1 - e2) - (2 * (e1 + de1x) - (e2 + de2x))
        PBEEXXbdifference = (2 * e1 - e2) - (2 * (e1 + de1xb) - (e2 + de2xb))
        print(PBEEXXdifference)
        print(PBEEXXbdifference)
        equal(PBEEXXdifference, 0.91, 0.005)
        equal(PBEEXXbdifference, 0.91, 0.005)



import numpy as np
from gpaw.utilities.ewald import Ewald


def test_utilities_ewald():
    verbose = True

    # References for Madelung constants with homogenous background charges:
    # M.P. Marder, "Condensed Matter Physics", J. Wiley, (2000),
    # scaled to the Wigner-Seitz radius.
    # References for Madelung constants of ionic crystals:
    # Y. Sakamoto, The J. of Chem. Phys., vol. 28, (1958), p. 164,
    # scaled to the nearest neighbor distance.

    if 1:  # fcc
        cell = np.array([[0., .5, .5], [.5, .0, .5], [.5, .5, .0]])
        basis = np.array([[0., 0., 0.]])
        charges = np.array([1.])
        r = np.array([0.0, 0.0, 0.0])
        charges = np.array([1])
        ewald = Ewald(cell)
        e = ewald.get_electrostatic_potential(r, basis, charges,
                                              excludefroml0=0)
        a_this = -e * ewald.get_wigner_seitz_radius(sum(charges))
        a_ref = 1.79175
        if verbose:
            print('Madelung energy, fcc:', a_this, a_this - a_ref)
        assert abs(a_this - a_ref) < 1e-5

    if 1:  # hcp
        a = 1.
        c = np.sqrt(8. / 3.) * a
        cell = np.array([[.5 * a, -0.5 * np.sqrt(3) * a, 0],
                         [.5 * a, 0.5 * np.sqrt(3) * a, 0], [0., 0., c]])
        basis = np.array([[.5 * a, .5 / np.sqrt(3) * a, 0.],
                          [.5 * a, -.5 / np.sqrt(3) * a, 0.5 * c]])
        r = np.array([.5 * a, .5 / np.sqrt(3) * a, 0.])
        charges = np.array([1., 1.])
        ewald = Ewald(cell)
        e = ewald.get_electrostatic_potential(r, basis, charges,
                                              excludefroml0=0)
        a_this = -e * ewald.get_wigner_seitz_radius(sum(charges))
        a_ref = 1.79168
        if verbose:
            print('Madelung energy, hcp:', a_this, a_this - a_ref)
        assert abs(a_this - a_ref) < 1e-5

    if 1:  # simple cubic
        cell = np.diag([1., 1., 1.])
        basis = np.array([[0., 0., 0.]])
        charges = np.array([1.])
        r = np.array([0.0, 0.0, 0.0])
        charges = np.array([1])
        ewald = Ewald(cell)
        e = ewald.get_electrostatic_potential(r, basis, charges,
                                              excludefroml0=0)
        a_this = -e * ewald.get_wigner_seitz_radius(1)
        a_ref = 1.76012
        if verbose:
            print('Madelung energy, sc:', a_this, a_this - a_ref)
        assert abs(a_this - a_ref) < 1e-5

    if 1:  # NaCl
        cell = np.array([[0., .5, .5], [.5, .0, .5], [.5, .5, .0]])
        basis = np.array([[0., 0., 0.], [.5, .5, .5]])
        r = np.array([0.0, 0.0, 0.0])
        charges = np.array([1, -1])
        ewald = Ewald(cell)
        e = ewald.get_electrostatic_potential(r, basis, charges,
                                              excludefroml0=0)
        a_this = -0.5 * e
        a_ref = 1.7475645946331822
        if verbose:
            print('Madelung constant, NaCl (B1):', a_this, a_this - a_ref)
        assert abs(a_this - a_ref) < 1e-13

    if 0:  # NaCl
        cell = np.diag((1., 1., 1.))
        basis = np.array([[0., 0., 0.], [.5, .5, .0],
                          [.5, .0, .5], [.0, .5, .5],
                          [.5, 0., 0.], [.0, .5, .0],
                          [.0, .0, .5], [.5, .5, .5]])
        r = np.array([0.0, 0.0, 0.0])
        charges = np.array([1, 1, 1, 1, -1, -1, -1, -1])
        ewald = Ewald(cell)
        e = ewald.get_electrostatic_potential(r, basis, charges,
                                              excludefroml0=0)
        a_this = -0.5 * e
        a_ref = 1.7475645946331822
        if verbose:
            print('Madelung constant, NaCl (B1):', a_this, a_this - a_ref)
        assert abs(a_this - a_ref) < 1e-13

    if 1:  # CsCl
        cell = np.diag((1., 1., 1.))
        basis = np.array([[0., 0., 0.], [.5, .5, .5]])
        r = np.array([0.0, 0.0, 0.0])
        charges = np.array([1, -1])
        ewald = Ewald(cell)
        e = ewald.get_electrostatic_potential(r, basis, charges,
                                              excludefroml0=0)
        a_this = -.5 * np.sqrt(3.) * e
        a_ref = 1.76267477307099
        if verbose:
            print('Madelung constant, CsCl (B2):', a_this, a_this - a_ref)
        assert abs(a_this - a_ref) < 1e-13

    if 1:  # Zincblende, cubic ZnS
        cell = np.array([[.0, .5, .5], [.5, .0, .5], [.5, .5, .0]])
        basis = np.array([[0., 0., 0.], [.25, .25, .25]])
        r = np.array([0.0, 0.0, 0.0])
        charges = np.array([1, -1])
        ewald = Ewald(cell)
        e = ewald.get_electrostatic_potential(r, basis, charges,
                                              excludefroml0=0)
        a_this = -0.25 * np.sqrt(3) * e
        a_ref = 1.63805505338879
        if verbose:
            print('Madelung constant, Zincblende (B3):',
                  a_this, a_this - a_ref)
        assert abs(a_this - a_ref) < 1e-13

    if 1:  # Ideal Wurtzite, ZnS, (B4)
        a = 1.
        c = np.sqrt(8. / 3.) * a
        u = 3. / 8.
        cell = np.array([[.5 * a, -0.5 * np.sqrt(3) * a, 0],
                         [.5 * a, 0.5 * np.sqrt(3) * a, 0], [0., 0., c]])
        basis = np.array([[.5 * a, .5 / np.sqrt(3) * a, 0.],
                          [.5 * a, -.5 / np.sqrt(3) * a, 0.5 * c],
                          [.5 * a, .5 / np.sqrt(3) * a, u * c],
                          [.5 * a, -.5 / np.sqrt(3) * a, (.5 + u) * c]])
        r = np.array([.5 * a, .5 / np.sqrt(3) * a, 0.])
        charges = np.array([1., 1., -1, -1])
        ewald = Ewald(cell)
        e = ewald.get_electrostatic_potential(r, basis, charges,
                                              excludefroml0=0)
        a_this = -1. * u * c * e
        a_ref = 1.64132162737
        if verbose:
            print('Madelung constant, Ideal Wurtzite (B4):', a_this,
                  a_this - a_ref)
        assert abs(a_this - a_ref) < 1e-11



"""This module implements a phonon perturbation."""

__all__ = ["PhononPerturbation"]

from math import pi
import numpy as np
import numpy.linalg as la

from gpaw.utilities import unpack
from gpaw.transformers import Transformer
from gpaw.lfc import LocalizedFunctionsCollection as LFC
from gpaw.dfpt.perturbation import Perturbation


class PhononPerturbation(Perturbation):
    """Implementation of a phonon perturbation.

    This class implements the change in the effective potential due to a
    displacement of an atom ``a`` in direction ``v`` with wave-vector ``q``.
    The action of the perturbing potential on a state vector is implemented in
    the ``apply`` member function.

    """
    def __init__(self, calc, kd, poisson_solver, dtype=float, **kwargs):
        """Store useful objects, e.g. lfc's for the various atomic functions.

        Depending on whether the system is periodic or finite, Poisson's
        equation is solved with FFT or multigrid techniques, respectively.

        Parameters
        ----------
        calc: Calculator
            Ground-state calculation.
        kd: KPointDescriptor
            Descriptor for the q-vectors of the dynamical matrix.

        """

        self.kd = kd
        self.dtype = dtype
        self.poisson = poisson_solver

        # Gamma wrt q-vector
        if self.kd.gamma:
            self.phase_cd = None
        else:
            assert self.kd.mynks == len(self.kd.ibzk_qc)

            self.phase_qcd = []
            sdisp_cd = calc.wfs.gd.sdisp_cd

            for q in range(self.kd.mynks):
                phase_cd = np.exp(2j * np.pi *
                                  sdisp_cd * self.kd.ibzk_qc[q, :, np.newaxis])
                self.phase_qcd.append(phase_cd)

        # Store grid-descriptors
        self.gd = calc.density.gd
        self.finegd = calc.density.finegd

        # Steal setups for the lfc's
        setups = calc.wfs.setups

        # Store projector coefficients
        self.dH_asp = calc.hamiltonian.dH_asp.copy()

        # Localized functions:
        # core corections
        self.nct = LFC(self.gd, [[setup.nct] for setup in setups],
                       integral=[setup.Nct for setup in setups],
                       dtype=self.dtype)
        # compensation charges
        # XXX what is the consequence of numerical errors in the integral ??
        self.ghat = LFC(self.finegd, [setup.ghat_l for setup in setups],
                        kd,
                        dtype=self.dtype)
        # self.ghat = LFC(self.finegd, [setup.ghat_l for setup in setups],
        #                 integral=sqrt(4 * pi), dtype=self.dtype)
        # vbar potential
        self.vbar = LFC(self.finegd, [[setup.vbar] for setup in setups],
                        kd,
                        dtype=self.dtype)

        # Expansion coefficients for the compensation charges
        self.Q_aL = calc.density.Q_aL.copy()

        # Grid transformer -- convert array from fine to coarse grid
        self.restrictor = Transformer(self.finegd,
                                      self.gd,
                                      nn=3,
                                      dtype=self.dtype)

        # Atom, cartesian coordinate and q-vector of the perturbation
        self.a = None
        self.v = None

        # Local q-vector index of the perturbation
        if self.kd.gamma:
            self.q = -1
        else:
            self.q = None

    def initialize(self, spos_ac):
        """Prepare the various attributes for a calculation."""

        # Set positions on LFC's
        self.nct.set_positions(spos_ac)
        self.ghat.set_positions(spos_ac)
        self.vbar.set_positions(spos_ac)

        if not self.kd.gamma:
            # Phase factor exp(iq.r) needed to obtian the periodic part of lfcs
            coor_vg = self.finegd.get_grid_point_coordinates()
            cell_cv = self.finegd.cell_cv
            # Convert to scaled coordinates
            scoor_cg = np.dot(la.inv(cell_cv), coor_vg.swapaxes(0, -2))
            scoor_cg = scoor_cg.swapaxes(1, -2)
            # Phase factor
            phase_qg = np.exp(
                2j * pi * np.dot(self.kd.ibzk_qc, scoor_cg.swapaxes(0, -2)))
            self.phase_qg = phase_qg.swapaxes(1, -2)

        # XXX To be removed from this class !!
        # Setup the Poisson solver -- to be used on the fine grid
        self.poisson.set_grid_descriptor(self.finegd)
        self.poisson.initialize()

    def set_q(self, q):
        """Set the index of the q-vector of the perturbation."""

        assert not self.kd.gamma, "Gamma-point calculation"

        self.q = q

        # Update phases and Poisson solver
        self.phase_cd = self.phase_qcd[q]
        self.poisson.set_q(self.kd.ibzk_qc[q])

        # Invalidate calculated quantities
        # - local part of perturbing potential
        self.v1_G = None

    def set_av(self, a, v):
        """Set atom and cartesian component of the perturbation.

        Parameters
        ----------
        a: int
            Index of the atom.
        v: int
            Cartesian component (0, 1 or 2) of the atomic displacement.

        """

        assert self.q is not None

        self.a = a
        self.v = v

        # Update derivative of local potential
        self.calculate_local_potential()

    def get_phase_cd(self):
        """Overwrite base class member function."""

        return self.phase_cd

    def has_q(self):
        """Overwrite base class member function."""

        return (not self.kd.gamma)

    def get_q(self):
        """Return q-vector."""

        assert not self.kd.gamma, "Gamma-point calculation."

        return self.kd.ibzk_qc[self.q]

    def solve_poisson(self, phi_g, rho_g):
        """Solve Poisson's equation for a Bloch-type charge distribution.

        More to come here ...

        Parameters
        ----------
        phi_g: GridDescriptor
            Grid for the solution of Poissons's equation.
        rho_g: GridDescriptor
            Grid with the charge distribution.

        """

        # assert phi_g.shape == rho_g.shape == self.phase_qg.shape[-3:], \
        #       ("Arrays have incompatible shapes.")
        assert self.q is not None, ("q-vector not set")

        # Gamma point calculation wrt the q-vector -> rho_g periodic
        if self.kd.gamma:
            # XXX NOTICE: solve_neutral
            self.poisson.solve_neutral(phi_g, rho_g)
        else:
            # Divide out the phase factor to get the periodic part
            rhot_g = rho_g / self.phase_qg[self.q]

            # Solve Poisson's equation for the periodic part of the potential
            # XXX NOTICE: solve_neutral
            self.poisson.solve_neutral(phi_g, rhot_g)

            # Return to Bloch form
            phi_g *= self.phase_qg[self.q]

    def calculate_local_potential(self):
        """Derivate of the local potential wrt an atomic displacements.

        The local part of the PAW potential has contributions from the
        compensation charges (``ghat``) and a spherical symmetric atomic
        potential (``vbar``).

        """

        assert self.a is not None
        assert self.v is not None
        assert self.q is not None

        a = self.a
        v = self.v

        # Expansion coefficients for the ghat functions
        Q_aL = self.ghat.dict(zero=True)
        # Remember sign convention for add_derivative method
        # And be sure not to change the dtype of the arrays by assigning values
        # to array elements.
        Q_aL[a][:] = -1 * self.Q_aL[a]

        # Grid for derivative of compensation charges
        ghat1_g = self.finegd.zeros(dtype=self.dtype)
        self.ghat.add_derivative(a, v, ghat1_g, c_axi=Q_aL, q=self.q)

        # Solve Poisson's eq. for the potential from the periodic part of the
        # compensation charge derivative
        v1_g = self.finegd.zeros(dtype=self.dtype)
        self.solve_poisson(v1_g, ghat1_g)

        # Store potential from the compensation charge
        self.vghat1_g = v1_g.copy()

        # Add derivative of vbar - sign convention in add_derivative method
        c_ai = self.vbar.dict(zero=True)
        c_ai[a][0] = -1.
        self.vbar.add_derivative(a, v, v1_g, c_axi=c_ai, q=self.q)

        # Store potential for the evaluation of the energy derivative
        self.v1_g = v1_g.copy()

        # Transfer to coarse grid
        v1_G = self.gd.zeros(dtype=self.dtype)
        self.restrictor.apply(v1_g, v1_G, phases=self.phase_cd)

        self.v1_G = v1_G

    def apply(self, psi_nG, y_nG, wfs, k, kplusq):
        """Apply perturbation to unperturbed wave-functions.

        Parameters
        ----------
        psi_nG: ndarray
            Set of grid vectors to which the perturbation is applied.
        y_nG: ndarray
            Output vectors.
        wfs: WaveFunctions
            Instance of class ``WaveFunctions``.
        k: int
            Index of the k-point for the vectors.
        kplusq: int
            Index of the k+q vector.

        """

        assert self.a is not None
        assert self.v is not None
        assert self.q is not None
        assert psi_nG.ndim in (3, 4)
        assert tuple(self.gd.n_c) == psi_nG.shape[-3:]

        if psi_nG.ndim == 3:
            y_nG += self.v1_G * psi_nG
        else:
            y_nG += self.v1_G[np.newaxis, :] * psi_nG

        self.apply_nonlocal_potential(psi_nG, y_nG, wfs, k, kplusq)

    def apply_nonlocal_potential(self, psi_nG, y_nG, wfs, k, kplusq):
        """Derivate of the non-local PAW potential wrt an atomic displacement.

        Parameters
        ----------
        k: int
            Index of the k-point being operated on.
        kplusq: int
            Index of the k+q vector.

        """

        assert self.a is not None
        assert self.v is not None
        assert psi_nG.ndim in (3, 4)
        assert tuple(self.gd.n_c) == psi_nG.shape[-3:]

        if psi_nG.ndim == 3:
            n = 1
        else:
            n = psi_nG.shape[0]

        a = self.a
        v = self.v

        P_ani = wfs.kpt_u[k].P_ani
        dP_aniv = wfs.kpt_u[k].dP_aniv
        pt = wfs.pt

        # < p_a^i | Psi_nk >
        P_ni = P_ani[a]
        # < dp_av^i | Psi_nk > - remember the sign convention of the derivative
        dP_ni = -1 * dP_aniv[a][..., v]

        # Expansion coefficients for the projectors on atom a
        dH_ii = unpack(self.dH_asp[a][0])

        # The derivative of the non-local PAW potential has two contributions
        # 1) Sum over projectors
        c_ni = np.dot(dP_ni, dH_ii)
        c_ani = pt.dict(shape=n, zero=True)
        c_ani[a] = c_ni
        # k+q !!
        pt.add(y_nG, c_ani, q=kplusq)

        # 2) Sum over derivatives of the projectors
        dc_ni = np.dot(P_ni, dH_ii)
        dc_ani = pt.dict(shape=n, zero=True)
        # Take care of sign of derivative in the coefficients
        dc_ani[a] = -1 * dc_ni
        # k+q !!
        pt.add_derivative(a, v, y_nG, dc_ani, q=kplusq)



import pytest
from ase import Atoms
from gpaw import GPAW
from gpaw.lcao.scissors import Scissors


@pytest.mark.later
def test_scissors():
    h2 = Atoms('2H2', [[0, 0, 0], [0, 0, 0.74],
                       [4, 0, 0], [4, 0, 0.74]])
    h2.center(vacuum=3.0)
    d = 1.0
    h2.calc = GPAW(mode='lcao',
                   basis='sz(dzp)',
                   eigensolver=Scissors([(-d, d, 2)]),
                   txt=None)
    h2.get_potential_energy()
    e1, e2, e3, e4 = h2.calc.get_eigenvalues()
    assert e2 - e1 == pytest.approx(d, abs=0.01)
    assert e4 - e3 == pytest.approx(d, abs=0.01)



"""Basic test of elph/DisplacementRunner

The main functionality is imported from ase.phonons.Displacement,
so won't test this here. Just test, whether the potential is reliably
extracted.
"""
import numpy as np
import pytest

from ase.build import bulk
from ase.utils.filecache import MultiFileJSONCache

from gpaw import GPAW
from gpaw.elph import DisplacementRunner
from gpaw.mpi import world


@pytest.mark.elph
def test_displacements(in_tmp_dir):
    # 2 atoms with one 1 valence electron each
    atoms = bulk('Li', crystalstructure='bcc', a=3.51, cubic=True)
    assert len(atoms) == 2
    # Note: usually we need to disable point group symmetry for displacements,
    # but we won't run actual displacements, so it saves time not to bother.
    calc = GPAW(mode='lcao',
                basis='sz(dzp)',
                # need more than one point for proper test
                kpts={'size': (2, 2, 2), 'gamma': True},
                # symmetry={'point_group': False},
                # parallel={'sl_auto': True, 'augment_grids':True,
                #          'band': 1, 'kpt': 1, 'domain': 4 },
                txt='elph_displacements.txt')
    atoms.calc = calc

    elph = DisplacementRunner(atoms=atoms, calc=atoms.calc,
                              supercell=(1, 1, 1), name='elph',
                              calculate_forces=False)
    elph.indices = []
    elph.run()
    del elph

    # read stuff back
    if world.rank == 0:
        cache = MultiFileJSONCache('elph')
        info = cache['info']
        assert info['supercell'] == [1, 1, 1]
        assert info['natom'] == 2
        assert info['delta'] == 0.01
        assert 'dr_version' in info

        # the following might change if defaults are changed
        Vt_G = cache['eq']['Vt_sG'][0]
        # print(Vt_G.shape, np.min(Vt_G), np.max(Vt_G))
        assert Vt_G.shape == (16, 16, 16)
        assert pytest.approx(np.min(Vt_G), abs=1e-6) == -0.5636114
        assert pytest.approx(np.max(Vt_G), abs=1e-6) == -0.1254635

        dH_all_asp = cache['eq']['dH_all_asp']
        # print(np.min(dH_all_asp[0][0]), np.max(dH_all_asp[0][0]))
        assert len(dH_all_asp) == 2
        dH_p = dH_all_asp[0][0]
        assert dH_p.shape == (15,)
        assert pytest.approx(np.min(dH_p), abs=1e-6) == -0.2249569
        assert pytest.approx(np.max(dH_p), abs=1e-6) == 0.0335627



import re

import numpy as np

from ase.utils import IOContext

from gpaw.lcaotddft.observer import TDDFTObserver


def convert_repr(r):
    # Integer
    try:
        return int(r)
    except ValueError:
        pass
    # Boolean
    b = {repr(False): False, repr(True): True}.get(r, None)
    if b is not None:
        return b
    # String
    s = r[1:-1]
    if repr(s) == r:
        return s
    raise RuntimeError('Unknown value: %s' % r)


class DipoleMomentWriter(TDDFTObserver):
    """Observer for writing time-dependent dipole moment data.

    The data is written in atomic units.

    The observer attaches to the TDDFT calculator during creation.

    Parameters
    ----------
    paw
        TDDFT calculator
    filename
        File for writing dipole moment data
    center
        If true, dipole moment is evaluated with the center of cell
        as the origin
    density
        Density type used for evaluating dipole moment.
        Use the default value for production calculations;
        others are for testing purposes.
        Possible values:
        ``'comp'``: ``rhot_g``,
        ``'pseudo'``: ``nt_sg``,
        ``'pseudocoarse'``: ``nt_sG``.
    force_new_file
        If true, new dipole moment file is created (erasing any existing one)
        even when restarting time propagation.
    interval
        Update interval. Value of 1 corresponds to evaluating and
        writing data after every propagation step.
    """
    version = 1

    def __init__(self, paw, filename: str, *,
                 center: bool = False,
                 density: str = 'comp',
                 force_new_file: bool = False,
                 interval: int = 1):
        TDDFTObserver.__init__(self, paw, interval)
        self.ioctx = IOContext()
        if paw.niter == 0 or force_new_file:
            # Initialize
            self.do_center = center
            self.density_type = density
            self.fd = self.ioctx.openfile(filename, comm=paw.world, mode='w')
            self._write_header(paw)
        else:
            # Read and continue
            self.read_header(filename)
            self.fd = self.ioctx.openfile(filename, comm=paw.world, mode='a')

    def _write(self, line):
        self.fd.write(line)
        self.fd.flush()

    def _write_header(self, paw):
        line = '# %s[version=%s]' % (self.__class__.__name__, self.version)
        line += ('(center=%s, density=%s)\n' %
                 (repr(self.do_center), repr(self.density_type)))
        line += ('# %15s %15s %22s %22s %22s\n' %
                 ('time', 'norm', 'dmx', 'dmy', 'dmz'))
        self._write(line)

    def read_header(self, filename):
        with open(filename, 'r') as f:
            line = f.readline()
        m_i = re.split("[^a-zA-Z0-9_=']+", line[2:])
        assert m_i.pop(0) == self.__class__.__name__
        for m in m_i:
            if '=' not in m:
                continue
            k, v = m.split('=')
            v = convert_repr(v)
            if k == 'version':
                assert v == self.version
                continue
            # Translate key
            k = {'center': 'do_center', 'density': 'density_type'}[k]
            setattr(self, k, v)

    def _write_init(self, paw):
        time = paw.time
        line = '# Start; Time = %.8lf\n' % time
        self._write(line)

    def _write_kick(self, paw):
        time = paw.time
        kick = paw.kick_strength
        line = '# Kick = [%22.12le, %22.12le, %22.12le]; ' % tuple(kick)
        line += 'Time = %.8lf\n' % time
        self._write(line)

    def calculate_dipole_moment(self, gd, rho_g, center=True):
        if center:
            center_v = 0.5 * gd.cell_cv.sum(0)
        else:
            center_v = np.zeros(3, dtype=float)
        r_vg = gd.get_grid_point_coordinates()
        dm_v = np.zeros(3, dtype=float)
        for v in range(3):
            dm_v[v] = - gd.integrate((r_vg[v] - center_v[v]) * rho_g)
        return dm_v

    def _write_dm(self, paw):
        time = paw.time
        density = paw.density
        if self.density_type == 'comp':
            rho_g = density.rhot_g
            gd = density.finegd
        elif self.density_type == 'pseudo':
            rho_g = density.nt_sg.sum(axis=0)
            gd = density.finegd
        elif self.density_type == 'pseudocoarse':
            rho_g = density.nt_sG.sum(axis=0)
            gd = density.gd
        else:
            raise RuntimeError('Unknown density type: %s' % self.density_type)

        norm = gd.integrate(rho_g)
        # dm = self.calculate_dipole_moment(gd, rho_g, center=self.do_center)
        dm = gd.calculate_dipole_moment(rho_g, center=self.do_center)
        if paw.hamiltonian.poisson.get_description() == 'FDTD+TDDFT':  # XXX
            dm += paw.hamiltonian.poisson.get_classical_dipole_moment()
        line = ('%20.8lf %20.8le %22.12le %22.12le %22.12le\n' %
                (time, norm, dm[0], dm[1], dm[2]))
        self._write(line)

    def _update(self, paw):
        if paw.action == 'init':
            self._write_init(paw)
        elif paw.action == 'kick':
            self._write_kick(paw)
        self._write_dm(paw)

    def __del__(self):
        self.ioctx.close()



import _gpaw
import numpy as np

from gpaw import debug
from gpaw.fd_operators import Laplace
from gpaw.transformers import Transformer
from gpaw.utilities.blas import axpy


class Preconditioner:
    def __init__(self, gd0, kin0, dtype=float, block=1):
        gd1 = gd0.coarsen()
        gd2 = gd1.coarsen()
        self.kin0 = kin0
        self.kin1 = Laplace(gd1, -0.5, 1, dtype)
        self.kin2 = Laplace(gd2, -0.5, 1, dtype)
        self.scratch0 = gd0.zeros((2, block), dtype, False)
        self.scratch1 = gd1.zeros((3, block), dtype, False)
        self.scratch2 = gd2.zeros((3, block), dtype, False)
        self.step = 0.66666666 / kin0.get_diagonal_element()

        self.restrictor_object0 = Transformer(gd0, gd1, 1, dtype)
        self.restrictor_object1 = Transformer(gd1, gd2, 1, dtype)
        self.interpolator_object2 = Transformer(gd2, gd1, 1, dtype)
        self.interpolator_object1 = Transformer(gd1, gd0, 1, dtype)
        self.restrictor0 = self.restrictor_object0.apply
        self.restrictor1 = self.restrictor_object1.apply
        self.interpolator2 = self.interpolator_object2.apply
        self.interpolator1 = self.interpolator_object1.apply
        self.use_c_precond = True

    def calculate_kinetic_energy(self, psit_xG, kpt):
        return None

    def __call__(self, residuals, kpt, ekin=None, out=None):
        if residuals.ndim == 3:
            if out is None:
                return self.__call__(residuals[np.newaxis], kpt)[0]
            return self.__call__(residuals[np.newaxis], kpt,
                                 out=out[np.newaxis])[0]

        nb = len(residuals)  # number of bands
        nb0 = self.scratch0.shape[1]
        if nb > nb0:
            assert out is not None
            for n1 in range(0, nb, nb0):
                self(residuals[n1:n1 + nb0], kpt, out=out[n1:n1 + nb0])
            return out

        phases = kpt.phase_cd
        step = self.step
        if out is None:
            d0, q0 = self.scratch0[:, :nb]
        else:
            d0 = out
            q0 = self.scratch0[0, :nb]
        r1, d1, q1 = self.scratch1[:, :nb]
        r2, d2, q2 = self.scratch2[:, :nb]
        if self.use_c_precond:
            transformers = [self.restrictor_object0.transformer,
                            self.restrictor_object1.transformer,
                            self.interpolator_object1.transformer,
                            self.interpolator_object2.transformer]
            if debug:
                # Unwrap wrapper:
                transformers = [getattr(t, 'transformer', t)
                                for t in transformers]
            _gpaw.fd_precond(*transformers,
                             self.kin0.operator, self.kin1.operator,
                             self.kin2.operator,
                             d0, q0, r1, d1, q1, r2, d2, q2,
                             residuals, -residuals, step, phases)
            return d0
        self.restrictor0(-residuals, r1, phases)
        d1[:] = 4 * step * r1
        self.kin1.apply(d1, q1, phases)
        q1 -= r1
        self.restrictor1(q1, r2, phases)
        d2 = 16 * step * r2
        self.kin2.apply(d2, q2, phases)
        q2 -= r2
        d2 -= 16 * step * q2
        self.interpolator2(d2, q1, phases)
        d1 -= q1
        self.kin1.apply(d1, q1, phases)
        q1 -= r1
        d1 -= 4 * step * q1
        self.interpolator1(-d1, d0, phases)
        self.kin0.apply(d0, q0, phases)
        q0 -= residuals
        axpy(-step, q0, d0)  # d0 -= step * q0
        d0 *= -1.0
        return d0
